package info.frostfs.sdk.services.impl;

import com.google.protobuf.ByteString;
import frostfs.ape.Types;
import frostfs.apemanager.APEManagerServiceGrpc;
import frostfs.apemanager.Service;
import info.frostfs.sdk.dto.ape.Chain;
import info.frostfs.sdk.jdo.ClientEnvironment;
import info.frostfs.sdk.jdo.parameters.CallContext;
import info.frostfs.sdk.jdo.parameters.ape.PrmApeChainAdd;
import info.frostfs.sdk.jdo.parameters.ape.PrmApeChainList;
import info.frostfs.sdk.jdo.parameters.ape.PrmApeChainRemove;
import info.frostfs.sdk.mappers.chain.ChainTargetMapper;
import info.frostfs.sdk.services.ApeManagerClient;
import info.frostfs.sdk.services.ContextAccessor;
import info.frostfs.sdk.tools.RequestConstructor;
import info.frostfs.sdk.tools.RequestSigner;
import info.frostfs.sdk.tools.Verifier;
import info.frostfs.sdk.tools.ape.RuleDeserializer;
import info.frostfs.sdk.tools.ape.RuleSerializer;

import java.util.List;
import java.util.stream.Collectors;

import static info.frostfs.sdk.utils.DeadLineUtil.deadLineAfter;
import static info.frostfs.sdk.utils.Validator.validate;

public class ApeManagerClientImpl extends ContextAccessor implements ApeManagerClient {
    private final APEManagerServiceGrpc.APEManagerServiceBlockingStub apeManagerServiceClient;

    public ApeManagerClientImpl(ClientEnvironment clientEnvironment) {
        super(clientEnvironment);
        this.apeManagerServiceClient = APEManagerServiceGrpc.newBlockingStub(getContext().getChannel());
    }

    @Override
    public byte[] addChain(PrmApeChainAdd args, CallContext ctx) {
        validate(args);

        var request = createAddChainRequest(args);

        var service = deadLineAfter(apeManagerServiceClient, ctx.getTimeout(), ctx.getTimeUnit());
        var response = service.addChain(request);

        Verifier.checkResponse(response);

        return response.getBody().getChainId().toByteArray();
    }

    @Override
    public void removeChain(PrmApeChainRemove args, CallContext ctx) {
        validate(args);

        var request = createRemoveChainRequest(args);

        var service = deadLineAfter(apeManagerServiceClient, ctx.getTimeout(), ctx.getTimeUnit());
        var response = service.removeChain(request);

        Verifier.checkResponse(response);
    }

    @Override
    public List<Chain> listChains(PrmApeChainList args, CallContext ctx) {
        validate(args);

        var request = createListChainsRequest(args);

        var service = deadLineAfter(apeManagerServiceClient, ctx.getTimeout(), ctx.getTimeUnit());
        var response = service.listChains(request);

        Verifier.checkResponse(response);

        return response.getBody().getChainsList().stream()
                .map(chain -> RuleDeserializer.deserialize(chain.getRaw().toByteArray()))
                .collect(Collectors.toList());
    }

    private Service.AddChainRequest createAddChainRequest(PrmApeChainAdd args) {
        var raw = RuleSerializer.serialize(args.getChain());

        var chainGrpc = Types.Chain.newBuilder()
                .setRaw(ByteString.copyFrom(raw))
                .build();
        var body = Service.AddChainRequest.Body.newBuilder()
                .setChain(chainGrpc)
                .setTarget(ChainTargetMapper.toGrpcMessage(args.getChainTarget()))
                .build();
        var request = Service.AddChainRequest.newBuilder()
                .setBody(body);

        RequestConstructor.addMetaHeader(request, args.getXHeaders());
        RequestSigner.sign(request, getContext().getKey());

        return request.build();
    }

    private Service.RemoveChainRequest createRemoveChainRequest(PrmApeChainRemove args) {
        var body = Service.RemoveChainRequest.Body.newBuilder()
                .setChainId(ByteString.copyFrom(args.getChainId()))
                .setTarget(ChainTargetMapper.toGrpcMessage(args.getChainTarget()))
                .build();
        var request = Service.RemoveChainRequest.newBuilder()
                .setBody(body);

        RequestConstructor.addMetaHeader(request, args.getXHeaders());
        RequestSigner.sign(request, getContext().getKey());

        return request.build();
    }

    private Service.ListChainsRequest createListChainsRequest(PrmApeChainList args) {
        var body = Service.ListChainsRequest.Body.newBuilder()
                .setTarget(ChainTargetMapper.toGrpcMessage(args.getChainTarget()))
                .build();
        var request = Service.ListChainsRequest.newBuilder()
                .setBody(body);

        RequestConstructor.addMetaHeader(request, args.getXHeaders());
        RequestSigner.sign(request, getContext().getKey());

        return request.build();
    }
}
