package info.frostfs.sdk.services.impl;

import frostfs.netmap.NetmapServiceGrpc;
import frostfs.netmap.Service;
import frostfs.netmap.Types;
import info.frostfs.sdk.dto.netmap.NetmapSnapshot;
import info.frostfs.sdk.dto.netmap.NodeInfo;
import info.frostfs.sdk.jdo.ClientEnvironment;
import info.frostfs.sdk.jdo.NetworkSettings;
import info.frostfs.sdk.jdo.parameters.CallContext;
import info.frostfs.sdk.mappers.netmap.NetmapSnapshotMapper;
import info.frostfs.sdk.mappers.netmap.NodeInfoMapper;
import info.frostfs.sdk.services.ContextAccessor;
import info.frostfs.sdk.services.NetmapClient;
import info.frostfs.sdk.tools.RequestConstructor;
import info.frostfs.sdk.tools.Verifier;

import java.nio.charset.StandardCharsets;

import static info.frostfs.sdk.tools.RequestSigner.sign;
import static info.frostfs.sdk.utils.DeadLineUtil.deadLineAfter;
import static java.util.Objects.nonNull;

public class NetmapClientImpl extends ContextAccessor implements NetmapClient {
    private final NetmapServiceGrpc.NetmapServiceBlockingStub netmapServiceClient;

    public NetmapClientImpl(ClientEnvironment clientEnvironment) {
        super(clientEnvironment);
        this.netmapServiceClient = NetmapServiceGrpc.newBlockingStub(getContext().getChannel());
    }

    private static boolean getBoolValue(byte[] bytes) {
        for (var byteValue : bytes) {
            if (byteValue != 0) {
                return true;
            }
        }

        return false;
    }

    private static long getLongValue(byte[] bytes) {
        long val = 0;
        for (var i = bytes.length - 1; i >= 0; i--) {
            val = (val << Byte.SIZE) + bytes[i];
        }

        return val;
    }

    private static void setNetworksParam(Types.NetworkConfig.Parameter param, NetworkSettings settings) {
        var key = new String(param.getKey().toByteArray(), StandardCharsets.UTF_8);

        var valueBytes = param.getValue().toByteArray();
        switch (key) {
            case "AuditFee":
                settings.setAuditFee(getLongValue(valueBytes));
                break;
            case "BasicIncomeRate":
                settings.setBasicIncomeRate(getLongValue(valueBytes));
                break;
            case "ContainerFee":
                settings.setContainerFee(getLongValue(valueBytes));
                break;
            case "ContainerAliasFee":
                settings.setContainerAliasFee(getLongValue(valueBytes));
                break;
            case "EpochDuration":
                settings.setEpochDuration(getLongValue(valueBytes));
                break;
            case "InnerRingCandidateFee":
                settings.setIRCandidateFee(getLongValue(valueBytes));
                break;
            case "MaxECDataCount":
                settings.setMaxECDataCount(getLongValue(valueBytes));
                break;
            case "MaxECParityCount":
                settings.setMaxECParityCount(getLongValue(valueBytes));
                break;
            case "MaxObjectSize":
                settings.setMaxObjectSize(getLongValue(valueBytes));
                break;
            case "WithdrawFee":
                settings.setWithdrawalFee(getLongValue(valueBytes));
                break;
            case "HomomorphicHashingDisabled":
                settings.setHomomorphicHashingDisabled(getBoolValue(valueBytes));
                break;
            case "MaintenanceModeAllowed":
                settings.setMaintenanceModeAllowed(getBoolValue(valueBytes));
                break;
            default:
                settings.getUnnamedSettings().put(key, valueBytes);
                break;
        }
    }

    @Override
    public NetworkSettings getNetworkSettings(CallContext ctx) {
        if (nonNull(getContext().getNetworkSettings())) {
            return getContext().getNetworkSettings();
        }

        var info = getNetworkInfo(ctx);

        var settings = new NetworkSettings();

        for (var param : info.getBody().getNetworkInfo().getNetworkConfig().getParametersList()) {
            setNetworksParam(param, settings);
        }

        getContext().setNetworkSettings(settings);

        return settings;
    }

    @Override
    public NodeInfo getLocalNodeInfo(CallContext ctx) {
        var request = Service.LocalNodeInfoRequest.newBuilder();

        RequestConstructor.addMetaHeader(request);
        sign(request, getContext().getKey());

        var service = deadLineAfter(netmapServiceClient, ctx.getTimeout(), ctx.getTimeUnit());
        var response = service.localNodeInfo(request.build());

        Verifier.checkResponse(response);

        return NodeInfoMapper.toModel(response.getBody());
    }

    public Service.NetworkInfoResponse getNetworkInfo(CallContext ctx) {
        var request = Service.NetworkInfoRequest.newBuilder();

        RequestConstructor.addMetaHeader(request);
        sign(request, getContext().getKey());

        var service = deadLineAfter(netmapServiceClient, ctx.getTimeout(), ctx.getTimeUnit());
        var response = service.networkInfo(request.build());

        Verifier.checkResponse(response);

        return response;
    }

    @Override
    public NetmapSnapshot getNetmapSnapshot(CallContext ctx) {
        var request = Service.NetmapSnapshotRequest.newBuilder();

        RequestConstructor.addMetaHeader(request);
        sign(request, getContext().getKey());

        var service = deadLineAfter(netmapServiceClient, ctx.getTimeout(), ctx.getTimeUnit());
        var response = service.netmapSnapshot(request.build());

        Verifier.checkResponse(response);

        return NetmapSnapshotMapper.toModel(response);
    }
}
