package info.frostfs.sdk.services.impl;

import com.google.common.collect.Iterables;
import com.google.protobuf.ByteString;
import frostfs.object.ObjectServiceGrpc;
import frostfs.object.Service;
import frostfs.refs.Types;
import info.frostfs.sdk.constants.AppConst;
import info.frostfs.sdk.dto.object.*;
import info.frostfs.sdk.dto.session.SessionToken;
import info.frostfs.sdk.enums.ObjectType;
import info.frostfs.sdk.exceptions.ProcessFrostFSException;
import info.frostfs.sdk.jdo.ClientEnvironment;
import info.frostfs.sdk.jdo.PutObjectResult;
import info.frostfs.sdk.jdo.parameters.CallContext;
import info.frostfs.sdk.jdo.parameters.object.*;
import info.frostfs.sdk.jdo.parameters.object.patch.PrmObjectPatch;
import info.frostfs.sdk.jdo.parameters.object.patch.PrmRangeGet;
import info.frostfs.sdk.jdo.parameters.object.patch.PrmRangeHashGet;
import info.frostfs.sdk.jdo.parameters.session.SessionContext;
import info.frostfs.sdk.jdo.result.ObjectHeaderResult;
import info.frostfs.sdk.mappers.container.ContainerIdMapper;
import info.frostfs.sdk.mappers.object.*;
import info.frostfs.sdk.mappers.object.patch.AddressMapper;
import info.frostfs.sdk.mappers.object.patch.RangeMapper;
import info.frostfs.sdk.services.ContextAccessor;
import info.frostfs.sdk.services.ObjectClient;
import info.frostfs.sdk.services.impl.rwhelper.*;
import info.frostfs.sdk.tools.RequestConstructor;
import info.frostfs.sdk.tools.Verifier;
import org.apache.commons.collections4.CollectionUtils;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;

import static info.frostfs.sdk.constants.ErrorConst.PROTO_MESSAGE_IS_EMPTY_TEMPLATE;
import static info.frostfs.sdk.tools.RequestSigner.sign;
import static info.frostfs.sdk.utils.DeadLineUtil.deadLineAfter;
import static info.frostfs.sdk.utils.Validator.validate;
import static java.util.Objects.isNull;
import static java.util.Objects.nonNull;

public class ObjectClientImpl extends ContextAccessor implements ObjectClient {
    private final ObjectServiceGrpc.ObjectServiceBlockingStub objectServiceBlockingClient;
    private final ObjectServiceGrpc.ObjectServiceStub objectServiceClient;
    private final ObjectToolsImpl objectToolsImpl;
    private final SessionToolsImpl sessionTools;

    public ObjectClientImpl(ClientEnvironment clientEnvironment) {
        super(clientEnvironment);
        this.objectServiceBlockingClient = ObjectServiceGrpc.newBlockingStub(getContext().getChannel());
        this.objectServiceClient = ObjectServiceGrpc.newStub(getContext().getChannel());
        this.objectToolsImpl = new ObjectToolsImpl(clientEnvironment);
        this.sessionTools = new SessionToolsImpl(clientEnvironment);
    }

    public SessionToken getOrCreateSession(SessionContext sessionContext, CallContext ctx) {
        return isNull(sessionContext.getSessionToken())
                ? sessionTools.getOrCreateSession(getContext(), ctx)
                : sessionContext.getSessionToken();
    }

    @Override
    public ObjectHeaderResult getObjectHead(PrmObjectHeadGet args, CallContext ctx) {
        validate(args);

        var request = createHeadRequest(args, ctx);

        var service = deadLineAfter(objectServiceBlockingClient, ctx.getTimeout(), ctx.getTimeUnit());
        var response = service.head(request);

        Verifier.checkResponse(response);

        return ObjectHeaderResult.builder()
                .headerInfo(ObjectHeaderMapper.toModel(response.getBody().getHeader().getHeader()))
                .splitInfo(SplitInfoMapper.toModel(response.getBody().getSplitInfo()))
                .build();
    }

    @Override
    public ObjectFrostFS getObject(PrmObjectGet args, CallContext ctx) {
        validate(args);

        var request = createGetRequest(args, ctx);

        return getObject(request, ctx);
    }

    @Override
    public void deleteObject(PrmObjectDelete args, CallContext ctx) {
        validate(args);

        var request = createDeleteRequest(args, ctx);

        var service = deadLineAfter(objectServiceBlockingClient, ctx.getTimeout(), ctx.getTimeUnit());
        var response = service.delete(request);

        Verifier.checkResponse(response);
    }

    @Override
    public Iterable<ObjectId> searchObjects(PrmObjectSearch args, CallContext ctx) {
        validate(args);

        var request = createSearchRequest(args, ctx);

        var objectsIds = searchObjects(request, ctx);

        return Iterables.transform(objectsIds, input -> new ObjectId(input.getValue().toByteArray()));
    }

    @Override
    public ObjectWriter putObject(PrmObjectPut args, CallContext ctx) {
        validate(args);

        return new ObjectWriter(getContext(), args, getUploadStream(args, ctx));
    }

    @Override
    public ObjectId putClientCutObject(PrmObjectClientCutPut args, CallContext ctx) {
        validate(args);

        var header = args.getObjectHeader();
        var fullLength = header.getPayloadLength() == 0 ? getStreamSize(args.getPayload()) : header.getPayloadLength();
        args.getPutObjectContext().setFullLength(fullLength);

        if (args.getPutObjectContext().getMaxObjectSizeCache() == 0) {
            var networkSettings = getContext().getFrostFSClient().getNetworkSettings(ctx);
            args.getPutObjectContext().setMaxObjectSizeCache(networkSettings.getMaxObjectSize().intValue());
        }

        var restBytes = fullLength - args.getPutObjectContext().getCurrentStreamPosition();
        var objectSize = restBytes > 0
                ? Math.min(args.getPutObjectContext().getMaxObjectSizeCache(), restBytes)
                : args.getPutObjectContext().getMaxObjectSizeCache();

        //define collection capacity
        var restPart = (restBytes % objectSize) > 0 ? 1 : 0;
        var objectsCount = fullLength > 0 ? (int) (restBytes / objectSize) + restPart : 0;

        List<ObjectId> sentObjectIds = new ArrayList<>(objectsCount);

        // keep attributes for the large object
        var attributes = args.getObjectHeader().getAttributes();

        Split split = new Split();
        args.getObjectHeader().setAttributes(new ArrayList<>());

        // send all parts except the last one as separate Objects
        while (restBytes > (long) args.getPutObjectContext().getMaxObjectSizeCache()) {
            var previous = CollectionUtils.isNotEmpty(sentObjectIds)
                    ? sentObjectIds.get(sentObjectIds.size() - 1)
                    : null;
            split.setPrevious(previous);
            args.getObjectHeader().setSplit(split);

            var result = putMultipartStreamObject(args, ctx);

            sentObjectIds.add(result.getObjectId());

            restBytes -= result.getObjectSize();
        }

        // send the last part and create linkObject
        if (CollectionUtils.isNotEmpty(sentObjectIds)) {
            var largeObjectHeader = new ObjectHeader(
                    header.getContainerId(), ObjectType.REGULAR, attributes, fullLength, header.getVersion()
            );
            largeObjectHeader.setOwnerId(header.getOwnerId());

            split.setParentHeader(largeObjectHeader);

            var result = putMultipartStreamObject(args, ctx);

            sentObjectIds.add(result.getObjectId());

            var linkObject = new LinkObject(header.getContainerId(), split.getSplitId(), largeObjectHeader);
            linkObject.addChildren(sentObjectIds);

            putSingleObject(new PrmObjectSinglePut(linkObject), ctx);

            return split.getParent();
        }

        // We are here if the payload is placed to one Object. It means no cut action, just simple PUT.
        var singlePartResult = putMultipartStreamObject(args, ctx);

        return singlePartResult.getObjectId();
    }

    @Override
    public ObjectId putSingleObject(PrmObjectSinglePut args, CallContext ctx) {
        var grpcObject = objectToolsImpl.createObject(args.getObjectFrostFS());
        var request = createPutSingleRequest(grpcObject, args, ctx);

        var service = deadLineAfter(objectServiceBlockingClient, ctx.getTimeout(), ctx.getTimeUnit());
        var response = service.putSingle(request);

        Verifier.checkResponse(response);

        return new ObjectId(grpcObject.getObjectId().getValue().toByteArray());
    }

    @Override
    public RangeReader getRange(PrmRangeGet args, CallContext ctx) {
        validate(args);

        var request = createGetRangeRequest(args, ctx);

        var service = deadLineAfter(objectServiceBlockingClient, ctx.getTimeout(), ctx.getTimeUnit());
        return new RangeReader(service.getRange(request));
    }

    @Override
    public byte[][] getRangeHash(PrmRangeHashGet args, CallContext ctx) {
        validate(args);

        var request = createGetRangeHashRequest(args, ctx);

        var service = deadLineAfter(objectServiceBlockingClient, ctx.getTimeout(), ctx.getTimeUnit());
        var response = service.getRangeHash(request);

        Verifier.checkResponse(response);

        return response.getBody().getHashListList().stream().map(ByteString::toByteArray).toArray(byte[][]::new);
    }

    @Override
    public ObjectId patchObject(PrmObjectPatch args, CallContext ctx) {
        validate(args);

        var request = createInitPatchRequest(args);
        var protoToken = RequestConstructor.createObjectTokenContext(
                getOrCreateSession(args, ctx),
                request.getBody().getAddress(),
                frostfs.session.Types.ObjectSessionContext.Verb.PATCH,
                getContext().getKey()
        );

        var currentPos = args.getRange().getOffset();
        var chunkSize = args.getMaxChunkLength();
        byte[] chunkBuffer = new byte[chunkSize];

        var service = deadLineAfter(objectServiceClient, ctx.getTimeout(), ctx.getTimeUnit());
        PatchStreamer writer = new PatchStreamer(service);

        var bytesCount = readNBytes(args.getPayload(), chunkBuffer, chunkSize);
        while (bytesCount > 0) {
            var range = Service.Range.newBuilder()
                    .setOffset(currentPos)
                    .setLength(bytesCount)
                    .build();

            Service.PatchRequest.Body.Patch.newBuilder()
                    .setChunk(ByteString.copyFrom(chunkBuffer, 0, bytesCount))
                    .setSourceRange(range)
                    .build();

            currentPos += bytesCount;

            RequestConstructor.addMetaHeader(request, args.getXHeaders(), protoToken);
            sign(request, getContext().getKey());

            writer.write(request.build());
            bytesCount = readNBytes(args.getPayload(), chunkBuffer, chunkSize);
        }

        var response = writer.complete();

        Verifier.checkResponse(response);

        return ObjectIdMapper.toModel(response.getBody().getObjectId());
    }

    private ObjectFrostFS getObject(Service.GetRequest request, CallContext ctx) {
        var reader = getObjectInit(request, ctx);

        var grpcObject = reader.readHeader();
        var modelObject = ObjectFrostFSMapper.toModel(grpcObject);

        modelObject.setObjectReader(reader);

        return modelObject;
    }

    private ObjectReaderImpl getObjectInit(Service.GetRequest initRequest, CallContext ctx) {
        if (initRequest.getSerializedSize() == 0) {
            throw new ProcessFrostFSException(
                    String.format(PROTO_MESSAGE_IS_EMPTY_TEMPLATE, initRequest.getClass().getName())
            );
        }

        var service = deadLineAfter(objectServiceBlockingClient, ctx.getTimeout(), ctx.getTimeUnit());
        return new ObjectReaderImpl(service.get(initRequest));
    }

    private PutObjectResult putMultipartStreamObject(PrmObjectClientCutPut args, CallContext ctx) {
        var chunkSize = args.getBufferMaxSize() > 0 ? args.getBufferMaxSize() : AppConst.OBJECT_CHUNK_SIZE;

        var restBytes =
                args.getPutObjectContext().getFullLength() - args.getPutObjectContext().getCurrentStreamPosition();

        chunkSize = (int) Math.min(restBytes, chunkSize);

        byte[] chunkBuffer = args.getCustomerBuffer() != null
                ? args.getCustomerBuffer()
                : new byte[chunkSize];//todo change to pool

        var sentBytes = 0;

        // 0 means no limit from client, so server side cut is performed
        var objectLimitSize = args.getPutObjectContext().getMaxObjectSizeCache();

        var stream = getUploadStream(args, ctx);

        while (objectLimitSize == 0 || sentBytes < objectLimitSize) {
            // send chunks limited to default or user's settings
            var bufferSize = objectLimitSize > 0 ? Math.min(objectLimitSize - sentBytes, chunkSize) : chunkSize;

            var bytesCount = readNBytes(args.getPayload(), chunkBuffer, bufferSize);

            if (bytesCount == 0) {
                break;
            }

            sentBytes += bytesCount;

            var body = Service.PutRequest.Body.newBuilder()
                    .setChunk(ByteString.copyFrom(chunkBuffer, 0, bytesCount))
                    .build();
            var chunkRequest = Service.PutRequest.newBuilder()
                    .setBody(body)
                    .clearVerifyHeader();

            RequestConstructor.addMetaHeader(chunkRequest, args.getXHeaders());
            sign(chunkRequest, getContext().getKey());

            stream.write(chunkRequest.build());
        }

        var response = stream.complete();
        Verifier.checkResponse(response);

        var objectId = new ObjectId(response.getBody().getObjectId().getValue().toByteArray());
        return new PutObjectResult(objectId, sentBytes);
    }

    private ObjectStreamer getUploadStream(PrmObjectPutBase args, CallContext ctx) {
        var header = args.getObjectHeader();

        header.setOwnerId(getContext().getOwnerId());
        header.setVersion(getContext().getVersion());

        var grpcHeader = ObjectHeaderMapper.toGrpcMessage(header);

        if (nonNull(header.getSplit())) {
            grpcHeader = objectToolsImpl.updateSplitValues(grpcHeader, header.getSplit());
        }

        var initRequest = createInitPutRequest(grpcHeader, args, ctx);

        return putObjectInit(initRequest, ctx);
    }

    private ObjectStreamer putObjectInit(Service.PutRequest initRequest, CallContext ctx) {
        if (initRequest.getSerializedSize() == 0) {
            throw new ProcessFrostFSException(
                    String.format(PROTO_MESSAGE_IS_EMPTY_TEMPLATE, initRequest.getClass().getName())
            );
        }

        var service = deadLineAfter(objectServiceClient, ctx.getTimeout(), ctx.getTimeUnit());
        ObjectStreamer writer = new ObjectStreamer(service);
        writer.write(initRequest);
        return writer;
    }

    private Iterable<Types.ObjectID> searchObjects(Service.SearchRequest request, CallContext ctx) {
        var reader = getSearchReader(request, ctx);
        var ids = reader.read();
        List<Types.ObjectID> result = new ArrayList<>();

        while (CollectionUtils.isNotEmpty(ids)) {
            result.addAll(ids);
            ids = reader.read();
        }

        return result;//todo return yield
    }

    private SearchReader getSearchReader(Service.SearchRequest initRequest, CallContext ctx) {
        if (initRequest.getSerializedSize() == 0) {
            throw new ProcessFrostFSException(
                    String.format(PROTO_MESSAGE_IS_EMPTY_TEMPLATE, initRequest.getClass().getName())
            );
        }

        var service = deadLineAfter(objectServiceBlockingClient, ctx.getTimeout(), ctx.getTimeUnit());
        return new SearchReader(service.search(initRequest));
    }

    private int readNBytes(InputStream inputStream, byte[] buffer, int size) {
        try {
            return inputStream.readNBytes(buffer, 0, size);
        } catch (IOException exp) {
            throw new ProcessFrostFSException(exp.getMessage());
        }
    }

    private long getStreamSize(InputStream inputStream) {
        try {
            return inputStream.available();
        } catch (IOException exp) {
            throw new ProcessFrostFSException(exp.getMessage());
        }
    }

    private Service.HeadRequest createHeadRequest(PrmObjectHeadGet args, CallContext ctx) {
        var address = Types.Address.newBuilder()
                .setContainerId(ContainerIdMapper.toGrpcMessage(args.getContainerId()))
                .setObjectId(ObjectIdMapper.toGrpcMessage(args.getObjectId()))
                .build();
        var body = Service.HeadRequest.Body.newBuilder()
                .setAddress(address)
                .setRaw(args.isRaw())
                .build();
        var request = Service.HeadRequest.newBuilder()
                .setBody(body);

        var sessionToken = getOrCreateSession(args, ctx);
        var protoToken = RequestConstructor.createObjectTokenContext(
                sessionToken,
                address,
                frostfs.session.Types.ObjectSessionContext.Verb.HEAD,
                getContext().getKey()
        );

        RequestConstructor.addMetaHeader(request, args.getXHeaders(), protoToken);
        sign(request, getContext().getKey());

        return request.build();
    }

    private Service.GetRequest createGetRequest(PrmObjectGet args, CallContext ctx) {
        var address = Types.Address.newBuilder()
                .setContainerId(ContainerIdMapper.toGrpcMessage(args.getContainerId()))
                .setObjectId(ObjectIdMapper.toGrpcMessage(args.getObjectId()))
                .build();
        var body = Service.GetRequest.Body.newBuilder()
                .setAddress(address)
                .build();
        var request = Service.GetRequest.newBuilder()
                .setBody(body);

        var sessionToken = getOrCreateSession(args, ctx);
        var protoToken = RequestConstructor.createObjectTokenContext(
                sessionToken,
                address,
                frostfs.session.Types.ObjectSessionContext.Verb.GET,
                getContext().getKey()
        );

        RequestConstructor.addMetaHeader(request, args.getXHeaders(), protoToken);
        sign(request, getContext().getKey());

        return request.build();
    }

    private Service.DeleteRequest createDeleteRequest(PrmObjectDelete args, CallContext ctx) {
        var address = Types.Address.newBuilder()
                .setContainerId(ContainerIdMapper.toGrpcMessage(args.getContainerId()))
                .setObjectId(ObjectIdMapper.toGrpcMessage(args.getObjectId()))
                .build();
        var body = Service.DeleteRequest.Body.newBuilder()
                .setAddress(address)
                .build();
        var request = Service.DeleteRequest.newBuilder()
                .setBody(body);

        var sessionToken = getOrCreateSession(args, ctx);
        var protoToken = RequestConstructor.createObjectTokenContext(
                sessionToken,
                address,
                frostfs.session.Types.ObjectSessionContext.Verb.DELETE,
                getContext().getKey()
        );

        RequestConstructor.addMetaHeader(request, args.getXHeaders(), protoToken);
        sign(request, getContext().getKey());

        return request.build();
    }

    private Service.SearchRequest createSearchRequest(PrmObjectSearch args, CallContext ctx) {
        var cid = ContainerIdMapper.toGrpcMessage(args.getContainerId());

        var address = Types.Address.newBuilder()
                .setContainerId(cid)
                .build();

        var body = Service.SearchRequest.Body.newBuilder()
                .setContainerId(cid)
                .setVersion(1);// TODO: clarify this param

        for (ObjectFilter<?> filter : args.getFilters()) {
            body.addFilters(ObjectFilterMapper.toGrpcMessage(filter));
        }

        var request = Service.SearchRequest.newBuilder()
                .setBody(body.build());

        var sessionToken = getOrCreateSession(args, ctx);
        var protoToken = RequestConstructor.createObjectTokenContext(
                sessionToken,
                address,
                frostfs.session.Types.ObjectSessionContext.Verb.SEARCH,
                getContext().getKey()
        );

        RequestConstructor.addMetaHeader(request, args.getXHeaders(), protoToken);
        sign(request, getContext().getKey());

        return request.build();
    }

    private Service.PutRequest createInitPutRequest(frostfs.object.Types.Header header,
                                                    PrmObjectPutBase args,
                                                    CallContext ctx) {
        var address = Types.Address.newBuilder()
                .setContainerId(header.getContainerId())
                .build();
        var init = Service.PutRequest.Body.Init.newBuilder()
                .setHeader(header)
                .build();
        var body = Service.PutRequest.Body.newBuilder()
                .setInit(init)
                .build();
        var request = Service.PutRequest.newBuilder()
                .setBody(body);

        var sessionToken = getOrCreateSession(args, ctx);
        var protoToken = RequestConstructor.createObjectTokenContext(
                sessionToken,
                address,
                frostfs.session.Types.ObjectSessionContext.Verb.PUT,
                getContext().getKey()
        );

        RequestConstructor.addMetaHeader(request, args.getXHeaders(), protoToken);
        sign(request, getContext().getKey());

        return request.build();
    }

    private Service.PutSingleRequest createPutSingleRequest(frostfs.object.Types.Object grpcObject,
                                                            PrmObjectSinglePut args,
                                                            CallContext ctx) {
        var address = Types.Address.newBuilder()
                .setContainerId(grpcObject.getHeader().getContainerId())
                .build();
        var body = Service.PutSingleRequest.Body.newBuilder()
                .setObject(grpcObject)
                .build();
        var request = Service.PutSingleRequest.newBuilder()
                .setBody(body);

        var sessionToken = getOrCreateSession(args, ctx);
        var protoToken = RequestConstructor.createObjectTokenContext(
                sessionToken,
                address,
                frostfs.session.Types.ObjectSessionContext.Verb.PUT,
                getContext().getKey()
        );

        RequestConstructor.addMetaHeader(request, args.getXHeaders(), protoToken);
        sign(request, getContext().getKey());

        return request.build();
    }

    private Service.GetRangeRequest createGetRangeRequest(PrmRangeGet args, CallContext ctx) {
        var address = Types.Address.newBuilder()
                .setContainerId(ContainerIdMapper.toGrpcMessage(args.getContainerId()))
                .setObjectId(ObjectIdMapper.toGrpcMessage(args.getObjectId()))
                .build();
        var body = Service.GetRangeRequest.Body.newBuilder()
                .setAddress(address)
                .setRange(RangeMapper.toGrpcMessage(args.getRange()))
                .setRaw(args.isRaw())
                .build();
        var request = Service.GetRangeRequest.newBuilder()
                .setBody(body);

        var sessionToken = getOrCreateSession(args, ctx);
        var protoToken = RequestConstructor.createObjectTokenContext(
                sessionToken,
                address,
                frostfs.session.Types.ObjectSessionContext.Verb.RANGE,
                getContext().getKey()
        );

        RequestConstructor.addMetaHeader(request, args.getXHeaders(), protoToken);
        sign(request, getContext().getKey());

        return request.build();
    }

    private Service.GetRangeHashRequest createGetRangeHashRequest(PrmRangeHashGet args, CallContext ctx) {
        var address = Types.Address.newBuilder()
                .setContainerId(ContainerIdMapper.toGrpcMessage(args.getContainerId()))
                .setObjectId(ObjectIdMapper.toGrpcMessage(args.getObjectId()))
                .build();
        var body = Service.GetRangeHashRequest.Body.newBuilder()
                .setAddress(address)
                .setType(Types.ChecksumType.SHA256)
                .setSalt(ByteString.copyFrom(args.getSalt()))
                .addAllRanges(RangeMapper.toGrpcMessages(args.getRanges()))
                .build();
        var request = Service.GetRangeHashRequest.newBuilder()
                .setBody(body);

        var sessionToken = getOrCreateSession(args, ctx);
        var protoToken = RequestConstructor.createObjectTokenContext(
                sessionToken,
                address,
                frostfs.session.Types.ObjectSessionContext.Verb.RANGEHASH,
                getContext().getKey()
        );

        RequestConstructor.addMetaHeader(request, args.getXHeaders(), protoToken);
        sign(request, getContext().getKey());

        return request.build();
    }

    private Service.PatchRequest.Builder createInitPatchRequest(PrmObjectPatch args) {
        var address = AddressMapper.toGrpcMessage(args.getAddress());
        var body = Service.PatchRequest.Body.newBuilder()
                .setAddress(address)
                .setReplaceAttributes(args.isReplaceAttributes())
                .addAllNewAttributes(ObjectAttributeMapper.toGrpcMessages(args.getNewAttributes()))
                .build();
        return Service.PatchRequest.newBuilder()
                .setBody(body);
    }
}
