package info.frostfs.sdk.services.impl;

import com.google.protobuf.ByteString;
import frostfs.object.Types;
import info.frostfs.sdk.dto.object.ObjectFrostFS;
import info.frostfs.sdk.dto.object.ObjectHeader;
import info.frostfs.sdk.dto.object.ObjectId;
import info.frostfs.sdk.dto.object.Split;
import info.frostfs.sdk.jdo.ClientEnvironment;
import info.frostfs.sdk.mappers.netmap.VersionMapper;
import info.frostfs.sdk.mappers.object.ObjectHeaderMapper;
import info.frostfs.sdk.mappers.object.ObjectIdMapper;
import info.frostfs.sdk.mappers.object.OwnerIdMapper;
import info.frostfs.sdk.services.ContextAccessor;
import info.frostfs.sdk.services.ToolsClient;
import org.apache.commons.collections4.CollectionUtils;

import java.util.stream.Collectors;

import static info.frostfs.sdk.Helper.getSha256;
import static info.frostfs.sdk.tools.RequestSigner.signData;
import static java.util.Objects.isNull;
import static java.util.Objects.nonNull;

public class ObjectToolsImpl extends ContextAccessor implements ToolsClient {
    public ObjectToolsImpl(ClientEnvironment context) {
        super(context);
    }

    private static frostfs.refs.Types.Checksum sha256Checksum(byte[] data) {
        return frostfs.refs.Types.Checksum.newBuilder()
                .setType(frostfs.refs.Types.ChecksumType.SHA256)
                .setSum(ByteString.copyFrom(getSha256(data)))
                .build();
    }

    @Override
    public ObjectId calculateObjectId(ObjectHeader header) {
        var grpcHeader = createHeader(header, new byte[]{});

        if (nonNull(header.getSplit())) {
            grpcHeader = updateSplitValues(grpcHeader, header.getSplit());
        }

        return ObjectIdMapper.toModel(
                frostfs.refs.Types.ObjectID.newBuilder().setValue(getSha256(grpcHeader)).build()
        );
    }

    public Types.Object createObject(ObjectFrostFS objectFrostFS) {
        objectFrostFS.getHeader().setOwnerId(getContext().getOwnerId());
        objectFrostFS.getHeader().setVersion(getContext().getVersion());
        objectFrostFS.getHeader().setPayloadLength(objectFrostFS.getPayload().length);

        var grpcHeader = ObjectHeaderMapper.toGrpcMessage(objectFrostFS.getHeader()).toBuilder()
                .setPayloadHash(sha256Checksum(objectFrostFS.getPayload()))
                .build();

        var split = objectFrostFS.getHeader().getSplit();
        if (nonNull(split)) {
            grpcHeader = updateSplitValues(grpcHeader, split);
        }

        var objectId = frostfs.refs.Types.ObjectID.newBuilder().setValue(getSha256(grpcHeader)).build();
        var sig = frostfs.refs.Types.Signature.newBuilder()
                .setKey(ByteString.copyFrom(getContext().getKey().getPublicKeyByte()))
                .setSign(ByteString.copyFrom(signData(getContext().getKey(), objectId.toByteArray())));
        return Types.Object.newBuilder()
                .setHeader(grpcHeader)
                .setObjectId(objectId)
                .setPayload(ByteString.copyFrom(objectFrostFS.getPayload()))
                .setSignature(sig)
                .build();
    }

    public Types.Header updateSplitValues(Types.Header grpcHeader, Split split) {
        if (isNull(split)) {
            return grpcHeader;
        }

        var grpcSplit = grpcHeader.getSplit().toBuilder()
                .setSplitId(ByteString.copyFrom(split.getSplitId().toBinary()));

        if (CollectionUtils.isNotEmpty(split.getChildren())) {
            var grpcChildren = split.getChildren().stream()
                    .map(ObjectIdMapper::toGrpcMessage)
                    .collect(Collectors.toList());
            grpcSplit.addAllChildren(grpcChildren);
        }

        if (nonNull(split.getParentHeader())) {
            var grpcParentHeader = createHeader(split.getParentHeader(), new byte[]{});
            var parentObjectId = frostfs.refs.Types.ObjectID.newBuilder().setValue(getSha256(grpcParentHeader)).build();
            var signature = frostfs.refs.Types.Signature.newBuilder()
                    .setKey(ByteString.copyFrom(getContext().getKey().getPublicKeyByte()))
                    .setSign(ByteString.copyFrom(signData(getContext().getKey(), parentObjectId.toByteArray())))
                    .build();


            grpcSplit
                    .setParent(parentObjectId)
                    .setParentHeader(grpcParentHeader)
                    .setParentSignature(signature);

            split.setParent(ObjectIdMapper.toModel(parentObjectId));
        }

        if (nonNull(split.getPrevious())) {
            grpcSplit.setPrevious(ObjectIdMapper.toGrpcMessage(split.getPrevious())).build();
        }

        return grpcHeader.toBuilder().setSplit(grpcSplit.build()).build();
    }

    private Types.Header createHeader(ObjectHeader header, byte[] payload) {
        var grpcHeader = ObjectHeaderMapper.toGrpcMessage(header).toBuilder()
                .setOwnerId(OwnerIdMapper.toGrpcMessage(getContext().getOwnerId()))
                .setVersion(VersionMapper.toGrpcMessage(getContext().getVersion()));

        if (payload != null) {
            grpcHeader.setPayloadHash(sha256Checksum(payload));
        }

        return grpcHeader.build();
    }
}
