package info.frostfs.sdk.services.impl;

import frostfs.session.Service;
import frostfs.session.SessionServiceGrpc;
import frostfs.session.Types;
import info.frostfs.sdk.dto.session.SessionToken;
import info.frostfs.sdk.jdo.ClientEnvironment;
import info.frostfs.sdk.jdo.parameters.CallContext;
import info.frostfs.sdk.jdo.parameters.session.PrmSessionCreate;
import info.frostfs.sdk.mappers.object.OwnerIdMapper;
import info.frostfs.sdk.mappers.session.SessionMapper;
import info.frostfs.sdk.services.ContextAccessor;
import info.frostfs.sdk.services.SessionClient;
import info.frostfs.sdk.tools.RequestConstructor;
import info.frostfs.sdk.tools.Verifier;

import static info.frostfs.sdk.tools.RequestSigner.sign;
import static info.frostfs.sdk.utils.DeadLineUtil.deadLineAfter;

public class SessionClientImpl extends ContextAccessor implements SessionClient {
    private final SessionServiceGrpc.SessionServiceBlockingStub serviceBlockingStub;

    public SessionClientImpl(ClientEnvironment clientEnvironment) {
        super(clientEnvironment);
        this.serviceBlockingStub = SessionServiceGrpc.newBlockingStub(getContext().getChannel());
    }

    @Override
    public SessionToken createSession(PrmSessionCreate args, CallContext ctx) {
        var sessionToken = createSessionInternal(args, ctx);
        var token = SessionMapper.serialize(sessionToken);
        return new SessionToken(token);
    }

    public Types.SessionToken createSessionInternal(PrmSessionCreate args, CallContext ctx) {
        var body = Service.CreateRequest.Body.newBuilder()
                .setOwnerId(OwnerIdMapper.toGrpcMessage(getContext().getOwnerId()))
                .setExpiration(args.getExpiration())
                .build();
        var request = Service.CreateRequest.newBuilder()
                .setBody(body);

        RequestConstructor.addMetaHeader(request);
        sign(request, getContext().getKey());

        return createSession(request.build(), ctx);
    }

    private Types.SessionToken createSession(Service.CreateRequest request, CallContext ctx) {
        var service = deadLineAfter(serviceBlockingStub, ctx.getTimeout(), ctx.getTimeUnit());
        var response = service.create(request);

        Verifier.checkResponse(response);

        var lifetime = Types.SessionToken.Body.TokenLifetime.newBuilder()
                .setExp(request.getBody().getExpiration())
                .setIat(response.getMetaHeader().getEpoch())
                .setNbf(response.getMetaHeader().getEpoch())
                .build();

        var body = Types.SessionToken.Body.newBuilder()
                .setId(response.getBody().getId())
                .setSessionKey(response.getBody().getSessionKey())
                .setOwnerId(request.getBody().getOwnerId())
                .setLifetime(lifetime)
                .build();

        return Types.SessionToken.newBuilder()
                .setBody(body)
                .build();
    }
}
