package info.frostfs.sdk.tools.ape;

import info.frostfs.sdk.dto.ape.*;
import info.frostfs.sdk.enums.ConditionKindType;
import info.frostfs.sdk.enums.ConditionType;
import info.frostfs.sdk.enums.RuleMatchType;
import info.frostfs.sdk.enums.RuleStatus;
import info.frostfs.sdk.exceptions.FrostFSException;
import info.frostfs.sdk.exceptions.ValidationFrostFSException;
import org.apache.commons.lang3.ArrayUtils;

import java.lang.reflect.Array;
import java.nio.charset.StandardCharsets;
import java.util.concurrent.atomic.AtomicInteger;

import static info.frostfs.sdk.constants.ErrorConst.*;
import static info.frostfs.sdk.constants.FieldConst.EMPTY_STRING;
import static info.frostfs.sdk.constants.RuleConst.*;

public class RuleDeserializer {
    private RuleDeserializer() {
    }

    public static Chain deserialize(byte[] data) {
        if (ArrayUtils.isEmpty(data)) {
            throw new ValidationFrostFSException(INPUT_PARAM_IS_MISSING);
        }

        AtomicInteger offset = new AtomicInteger(0);
        Chain chain = new Chain();

        var version = uInt8Unmarshal(data, offset);
        if (version != VERSION) {
            throw new FrostFSException(String.format(UNSUPPORTED_MARSHALLER_VERSION_TEMPLATE, version));
        }

        var chainVersion = uInt8Unmarshal(data, offset);
        if (chainVersion != CHAIN_MARSHAL_VERSION) {
            throw new FrostFSException(String.format(UNSUPPORTED_CHAIN_VERSION_TEMPLATE, chainVersion));
        }

        chain.setId(sliceUnmarshal(data, offset, Byte.class, RuleDeserializer::uInt8Unmarshal));
        chain.setRules(sliceUnmarshal(data, offset, Rule.class, RuleDeserializer::unmarshalRule));
        chain.setMatchType(RuleMatchType.get(uInt8Unmarshal(data, offset)));

        verifyUnmarshal(data, offset);

        return chain;
    }

    private static Byte uInt8Unmarshal(byte[] buf, AtomicInteger offset) {
        if (buf.length - offset.get() < 1) {
            throw new FrostFSException(
                    String.format(BYTES_ARE_OVER_FOR_DESERIALIZE_TEMPLATE, Byte.class.getName(), offset.get()));
        }

        return buf[offset.getAndAdd(1)];
    }

    public static long varInt(byte[] buf, AtomicInteger offset) {
        long ux = uVarInt(buf, offset); // ok to continue in presence of error
        long x = ux >> 1;
        if ((ux & 1) != 0) {
            x = ~x;
        }
        return x;
    }


    public static long uVarInt(byte[] buf, AtomicInteger offset) {
        long x = 0;
        int s = 0;

        for (int i = offset.get(); i < buf.length; i++) {
            long b = buf[i];
            if (i == MAX_VAR_INT_LENGTH) {
                offset.set(-(i + 1));
                return 0; // overflow
            }
            if (b >= 0) {
                if (i == MAX_VAR_INT_LENGTH - 1 && b > 1) {
                    offset.set(-(i + 1));
                    return 0; // overflow
                }
                offset.set(i + 1);
                return x | (b << s);
            }
            x |= (b & OFFSET127) << s;
            s += UNSIGNED_SERIALIZE_SIZE;
        }
        offset.set(0);
        return 0;
    }

    private static <T> T[] sliceUnmarshal(byte[] buf,
                                          AtomicInteger offset,
                                          Class<T> clazz,
                                          UnmarshalFunction<T> unmarshalT) {
        var size = (int) varInt(buf, offset);
        if (size == NULL_SLICE) {
            return null;
        }

        if (size > MAX_SLICE_LENGTH) {
            throw new ValidationFrostFSException(String.format(SLICE_IS_TOO_BIG_TEMPLATE, size));
        }

        if (size < 0) {
            throw new ValidationFrostFSException(String.format(SLICE_SIZE_IS_INVALID_TEMPLATE, size));
        }

        T[] result = (T[]) Array.newInstance(clazz, size);
        for (int i = 0; i < result.length; i++) {
            result[i] = unmarshalT.unmarshal(buf, offset);
        }

        return result;
    }

    private static boolean boolUnmarshal(byte[] buf, AtomicInteger offset) {
        return uInt8Unmarshal(buf, offset) == BYTE_TRUE;
    }

    private static long int64Unmarshal(byte[] buf, AtomicInteger offset) {
        if (buf.length - offset.get() < Long.BYTES) {
            throw new ValidationFrostFSException(
                    String.format(BYTES_ARE_OVER_FOR_DESERIALIZE_TEMPLATE, Long.class.getName(), offset.get())
            );
        }

        return varInt(buf, offset);
    }

    private static String stringUnmarshal(byte[] buf, AtomicInteger offset) {
        int size = (int) int64Unmarshal(buf, offset);
        if (size == 0) {
            return EMPTY_STRING;
        }

        if (size > MAX_SLICE_LENGTH) {
            throw new ValidationFrostFSException(String.format(STRING_IS_TOO_BIG_TEMPLATE, size));
        }

        if (size < 0) {
            throw new ValidationFrostFSException(String.format(STRING_SIZE_IS_INVALID_TEMPLATE, size));
        }

        if (buf.length - offset.get() < size) {
            throw new ValidationFrostFSException(
                    String.format(BYTES_ARE_OVER_FOR_DESERIALIZE_TEMPLATE, String.class.getName(), offset.get())
            );
        }

        return new String(buf, offset.getAndAdd(size), size, StandardCharsets.UTF_8);
    }

    private static Actions unmarshalActions(byte[] buf, AtomicInteger offset) {
        Actions actions = new Actions();
        actions.setInverted(boolUnmarshal(buf, offset));
        actions.setNames(sliceUnmarshal(buf, offset, String.class, RuleDeserializer::stringUnmarshal));
        return actions;
    }

    private static Condition unmarshalCondition(byte[] buf, AtomicInteger offset) {
        Condition condition = new Condition();
        condition.setOp(ConditionType.get(uInt8Unmarshal(buf, offset)));
        condition.setKind(ConditionKindType.get(uInt8Unmarshal(buf, offset)));
        condition.setKey(stringUnmarshal(buf, offset));
        condition.setValue(stringUnmarshal(buf, offset));

        return condition;
    }

    private static Rule unmarshalRule(byte[] buf, AtomicInteger offset) {
        Rule rule = new Rule();
        rule.setStatus(RuleStatus.get(uInt8Unmarshal(buf, offset)));
        rule.setActions(unmarshalActions(buf, offset));
        rule.setResources(unmarshalResources(buf, offset));
        rule.setAny(boolUnmarshal(buf, offset));
        rule.setConditions(sliceUnmarshal(buf, offset, Condition.class, RuleDeserializer::unmarshalCondition));

        return rule;
    }

    private static Resources unmarshalResources(byte[] buf, AtomicInteger offset) {
        Resources resources = new Resources();
        resources.setInverted(boolUnmarshal(buf, offset));
        resources.setNames(sliceUnmarshal(buf, offset, String.class, RuleDeserializer::stringUnmarshal));

        return resources;
    }

    private static void verifyUnmarshal(byte[] buf, AtomicInteger offset) {
        if (buf.length != offset.get()) {
            throw new ValidationFrostFSException(UNMARSHAL_SIZE_DIFFERS);
        }
    }
}
