package info.frostfs.sdk.tools.ape;

import info.frostfs.sdk.dto.ape.*;
import info.frostfs.sdk.exceptions.ValidationFrostFSException;
import org.apache.commons.lang3.StringUtils;

import java.nio.charset.StandardCharsets;
import java.util.function.Function;

import static info.frostfs.sdk.constants.ErrorConst.*;
import static info.frostfs.sdk.constants.RuleConst.*;
import static java.util.Objects.isNull;
import static java.util.Objects.nonNull;

public class RuleSerializer {
    private RuleSerializer() {
    }

    public static byte[] serialize(Chain chain) {
        if (isNull(chain)) {
            throw new ValidationFrostFSException(String.format(INPUT_PARAM_IS_MISSING_TEMPLATE, Chain.class.getName()));
        }

        int s = U_INT_8_SIZE // Marshaller version
                + U_INT_8_SIZE // Chain version
                + sliceSize(chain.getId(), b -> BYTE_SIZE)
                + sliceSize(chain.getRules(), RuleSerializer::ruleSize)
                + U_INT_8_SIZE; // MatchType

        byte[] buf = new byte[s];

        int offset = uInt8Marshal(buf, 0, VERSION);
        offset = uInt8Marshal(buf, offset, (byte) CHAIN_MARSHAL_VERSION);
        offset = sliceMarshal(buf, offset, chain.getId(), RuleSerializer::byteMarshal);
        offset = sliceMarshal(buf, offset, chain.getRules(), RuleSerializer::marshalRule);
        offset = uInt8Marshal(buf, offset, (byte) chain.getMatchType().value);

        verifyMarshal(buf, offset);

        return buf;
    }

    private static <T> int sliceSize(T[] slice, Function<T, Integer> sizeOf) {
        if (isNull(slice)) {
            return NULL_SLICE_SIZE;
        }

        // Assuming int64Size is the size of the slice
        int size = int64Size(slice.length);
        for (T v : slice) {
            size += sizeOf.apply(v);
        }

        return size;
    }

    /*
     * https://cs.opensource.google/go/go/+/master:src/encoding/binary/varint.go;l=92;drc=dac9b9ddbd5160c5f4552410f5f828
     * 1bd5eed38c
     *
     * and
     *
     * https://cs.opensource.google/go/go/+/master:src/encoding/binary/varint.go;l=41;drc=dac9b9ddbd5160c5f4552410f5f828
     * 1bd5eed38c
     * */
    private static int int64Size(long value) {
        long ux = value << 1;
        if (value < 0) {
            ux = ~ux;
        }

        int size = 0;
        while (ux >= OFFSET128) {
            size++;
            ux >>>= UNSIGNED_SERIALIZE_SIZE;
        }

        return size + 1;
    }

    private static int stringSize(String s) {
        int len = nonNull(s) ? s.length() : 0;
        return int64Size(len) + len;
    }

    private static int actionsSize(Actions action) {
        return BOOL_SIZE // Inverted
                + (nonNull(action) ? sliceSize(action.getNames(), RuleSerializer::stringSize) : 0);
    }

    private static int resourcesSize(Resources resource) {
        return BOOL_SIZE // Inverted
                + (nonNull(resource) ? sliceSize(resource.getNames(), RuleSerializer::stringSize) : 0);
    }

    private static int conditionSize(Condition condition) {
        if (isNull(condition)) {
            throw new ValidationFrostFSException(String.format(REQUIRED_FIELD_TEMPLATE, Rule.class.getName()));
        }

        return BYTE_SIZE // Op
                + BYTE_SIZE // Object
                + stringSize(condition.getKey())
                + stringSize(condition.getValue());
    }

    private static int ruleSize(Rule rule) {
        if (isNull(rule)) {
            throw new ValidationFrostFSException(String.format(REQUIRED_FIELD_TEMPLATE, Rule.class.getName()));
        }

        return BYTE_SIZE // Status
                + actionsSize(rule.getActions())
                + resourcesSize(rule.getResources())
                + BOOL_SIZE // Any
                + sliceSize(rule.getConditions(), RuleSerializer::conditionSize);
    }

    private static int uInt8Marshal(byte[] buf, int offset, byte value) {
        if (buf.length - offset < 1) {
            throw new ValidationFrostFSException(
                    String.format(BYTES_ARE_OVER_FOR_SERIALIZE_TEMPLATE, Byte.class.getName(), 1)
            );
        }

        buf[offset] = value;

        return offset + 1;
    }

    private static int byteMarshal(byte[] buf, int offset, byte value) {
        return uInt8Marshal(buf, offset, value);
    }

    // putVarInt encodes an int64 into buf and returns the number of bytes written.
    private static int putVarInt(byte[] buf, int offset, long x) {
        long ux = x << 1;
        if (x < 0) {
            ux = ~ux;
        }

        return putUVarInt(buf, offset, ux);
    }

    private static int putUVarInt(byte[] buf, int offset, long x) {
        while (x >= OFFSET128) {
            buf[offset] = (byte) (x | OFFSET128);
            x >>>= UNSIGNED_SERIALIZE_SIZE;
            offset++;
        }
        buf[offset] = (byte) x;
        return offset + 1;
    }

    private static int int64Marshal(byte[] buf, int offset, long v) {
        var size = int64Size(v);
        if (buf.length - offset < size) {
            throw new ValidationFrostFSException(
                    String.format(BYTES_ARE_OVER_FOR_SERIALIZE_TEMPLATE, Long.class.getName(), size)
            );
        }

        return putVarInt(buf, offset, v);
    }

    private static <T> int sliceMarshal(byte[] buf, int offset, T[] slice, MarshalFunction<T> marshalT) {
        if (isNull(slice)) {
            return int64Marshal(buf, offset, NULL_SLICE);
        }

        if (slice.length > MAX_SLICE_LENGTH) {
            throw new ValidationFrostFSException(String.format(SLICE_IS_TOO_BIG_TEMPLATE, slice.length));
        }

        offset = int64Marshal(buf, offset, slice.length);
        for (T v : slice) {
            offset = marshalT.marshal(buf, offset, v);
        }

        return offset;
    }

    private static int boolMarshal(byte[] buf, int offset, boolean value) {
        return uInt8Marshal(buf, offset, value ? BYTE_TRUE : BYTE_FALSE);
    }

    private static int stringMarshal(byte[] buf, int offset, String value) {
        if (StringUtils.isBlank(value)) {
            throw new ValidationFrostFSException(STRING_IS_BLANK);
        }

        if (value.length() > MAX_SLICE_LENGTH) {
            throw new ValidationFrostFSException(String.format(STRING_IS_TOO_BIG_TEMPLATE, value.length()));
        }

        if (buf.length - offset < int64Size(value.length()) + value.length()) {
            throw new ValidationFrostFSException(
                    String.format(BYTES_ARE_OVER_FOR_SERIALIZE_TEMPLATE, String.class.getName(), value.length())
            );
        }

        offset = int64Marshal(buf, offset, value.length());
        if (value.isEmpty()) {
            return offset;
        }

        byte[] stringBytes = value.getBytes(StandardCharsets.UTF_8);

        // Copy exactly value.length() bytes as in the original code
        System.arraycopy(stringBytes, 0, buf, offset, value.length());
        return offset + value.length();
    }

    private static int marshalActions(byte[] buf, int offset, Actions action) {
        if (isNull(action)) {
            throw new ValidationFrostFSException(String.format(REQUIRED_FIELD_TEMPLATE, Actions.class.getName()));
        }

        offset = boolMarshal(buf, offset, action.isInverted());
        return sliceMarshal(buf, offset, action.getNames(), RuleSerializer::stringMarshal);
    }

    private static int marshalCondition(byte[] buf, int offset, Condition condition) {
        if (isNull(condition)) {
            throw new ValidationFrostFSException(String.format(REQUIRED_FIELD_TEMPLATE, Condition.class.getName()));
        }

        offset = byteMarshal(buf, offset, (byte) condition.getOp().value);
        offset = byteMarshal(buf, offset, (byte) condition.getKind().value);
        offset = stringMarshal(buf, offset, condition.getKey());
        return stringMarshal(buf, offset, condition.getValue());
    }

    private static int marshalRule(byte[] buf, int offset, Rule rule) {
        if (isNull(rule)) {
            throw new ValidationFrostFSException(String.format(REQUIRED_FIELD_TEMPLATE, Rule.class.getName()));
        }

        offset = byteMarshal(buf, offset, (byte) rule.getStatus().value);
        offset = marshalActions(buf, offset, rule.getActions());
        offset = marshalResources(buf, offset, rule.getResources());
        offset = boolMarshal(buf, offset, rule.isAny());
        return sliceMarshal(buf, offset, rule.getConditions(), RuleSerializer::marshalCondition);
    }

    private static int marshalResources(byte[] buf, int offset, Resources resources) {
        if (isNull(resources)) {
            throw new ValidationFrostFSException(String.format(REQUIRED_FIELD_TEMPLATE, Resources.class.getName()));
        }

        offset = boolMarshal(buf, offset, resources.isInverted());
        return sliceMarshal(buf, offset, resources.getNames(), RuleSerializer::stringMarshal);
    }

    private static void verifyMarshal(byte[] buf, int lastOffset) {
        if (buf.length != lastOffset) {
            throw new ValidationFrostFSException(MARSHAL_SIZE_DIFFERS);
        }
    }
}
