package info.frostfs.sdk.utils;

import info.frostfs.sdk.annotations.*;
import info.frostfs.sdk.constants.ErrorConst;
import info.frostfs.sdk.exceptions.ProcessFrostFSException;
import info.frostfs.sdk.exceptions.ValidationFrostFSException;
import org.apache.commons.lang3.StringUtils;

import java.lang.reflect.Field;

import static info.frostfs.sdk.constants.ErrorConst.FIELDS_DELIMITER_COMMA;
import static java.util.Objects.isNull;

public class Validator {
    private Validator() {
    }

    public static <T> void validate(T object) {
        StringBuilder errorMessage = new StringBuilder();
        process(object, errorMessage);

        if (errorMessage.length() != 0) {
            errorMessage.insert(0, System.lineSeparator());
            throw new ValidationFrostFSException(errorMessage.toString());
        }
    }


    private static <T> void process(T object, StringBuilder errorMessage) {
        if (isNull(object)) {
            errorMessage.append(ErrorConst.OBJECT_IS_NULL);
            return;
        }

        Class<?> clazz = object.getClass();

        if (clazz.isAnnotationPresent(ComplexAtLeastOneIsFilled.class)) {
            processComplexAtLeastOneIsFilled(object, clazz, errorMessage);
        }

        if (clazz.isAnnotationPresent(AtLeastOneIsFilled.class)) {
            processAtLeastOneIsFilled(object, clazz, errorMessage);
        }

        for (Field field : clazz.getDeclaredFields()) {
            field.setAccessible(true);

            if (field.isAnnotationPresent(Validate.class)) {
                processValidate(object, errorMessage, field);
            }

            if (field.isAnnotationPresent(NotNull.class)) {
                processNotNull(object, errorMessage, field);
            }

            if (field.isAnnotationPresent(NotBlank.class)) {
                processNotBlank(object, errorMessage, field);
            }
        }
    }

    private static <T> void processNotNull(T object, StringBuilder errorMessage, Field field) {
        if (isNull(getFieldValue(object, field))) {
            errorMessage
                    .append(String.format(ErrorConst.REQUIRED_FIELD_TEMPLATE, field.getName()))
                    .append(System.lineSeparator());
        }
    }

    private static <T> void processNotBlank(T object, StringBuilder errorMessage, Field field) {
        var value = getFieldValue(object, field);
        if (isNull(value) || (value instanceof String && StringUtils.isBlank((String) value))) {
            errorMessage
                    .append(String.format(ErrorConst.REQUIRED_FIELD_TEMPLATE, field.getName()))
                    .append(System.lineSeparator());
        }
    }

    private static <T> void processValidate(T object, StringBuilder errorMessage, Field field) {
        if (isNull(object)) {
            return;
        }

        process(getFieldValue(object, field), errorMessage);
    }

    private static <T> void processComplexAtLeastOneIsFilled(T object, Class<?> clazz, StringBuilder errorMessage) {
        var annotation = clazz.getAnnotation(ComplexAtLeastOneIsFilled.class);
        for (AtLeastOneIsFilled value : annotation.value()) {
            processAtLeastOneIsFilled(object, clazz, errorMessage, value);
        }
    }

    private static <T> void processAtLeastOneIsFilled(T object, Class<?> clazz, StringBuilder errorMessage) {
        var annotation = clazz.getAnnotation(AtLeastOneIsFilled.class);
        processAtLeastOneIsFilled(object, clazz, errorMessage, annotation);
    }

    private static <T> void processAtLeastOneIsFilled(T object,
                                                      Class<?> clazz,
                                                      StringBuilder errorMessage,
                                                      AtLeastOneIsFilled annotation) {
        var emptyFieldsCount = 0;
        for (String fieldName : annotation.fields()) {
            var field = getField(clazz, fieldName);
            field.setAccessible(true);

            var value = getFieldValue(object, field);
            if (isNull(value) || (value instanceof String && StringUtils.isBlank((String) value))) {
                emptyFieldsCount++;
            }
        }

        if (emptyFieldsCount == annotation.fields().length) {
            errorMessage
                    .append(String.format(
                            annotation.message(),
                            String.join(FIELDS_DELIMITER_COMMA, annotation.fields())
                    ))
                    .append(System.lineSeparator());
        }
    }


    private static <T> Object getFieldValue(T object, Field field) {
        try {
            return field.get(object);
        } catch (IllegalAccessException e) {
            throw new ProcessFrostFSException(e);
        }
    }

    private static Field getField(Class<?> clazz, String fieldName) {
        try {
            return clazz.getDeclaredField(fieldName);
        } catch (NoSuchFieldException e) {
            throw new ProcessFrostFSException(e);
        }
    }
}
