package info.frostfs.sdk.services;

import frostfs.accounting.AccountingServiceGrpc;
import frostfs.accounting.Service;
import info.frostfs.sdk.dto.object.OwnerId;
import info.frostfs.sdk.jdo.ClientEnvironment;
import info.frostfs.sdk.jdo.parameters.CallContext;
import info.frostfs.sdk.services.impl.AccountingClientImpl;
import info.frostfs.sdk.testgenerator.AccountingGenerator;
import info.frostfs.sdk.tools.RequestConstructor;
import info.frostfs.sdk.tools.RequestSigner;
import info.frostfs.sdk.tools.Verifier;
import io.grpc.Channel;
import io.neow3j.crypto.Base58;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.junit.platform.commons.util.ReflectionUtils;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.MockedStatic;
import org.mockito.Mockito;
import org.mockito.junit.jupiter.MockitoExtension;

import java.lang.reflect.Field;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.mockito.Mockito.*;

@ExtendWith(MockitoExtension.class)
class AccountingClientTest {
    private static final String OWNER_ID = "NVxUSpEEJzYXZZtUs18PrJTD9QZkLLNQ8S";

    private AccountingClientImpl accountingClient;

    @Mock
    private AccountingServiceGrpc.AccountingServiceBlockingStub AccountingServiceClient;
    @Mock
    private ClientEnvironment clientEnvironment;
    @Mock
    private Channel channel;

    private MockedStatic<Verifier> verifierMock;
    private MockedStatic<RequestConstructor> requestConstructorMock;
    private MockedStatic<RequestSigner> requestSignerMock;

    @BeforeEach
    void setUp() throws IllegalAccessException {
        when(clientEnvironment.getChannel()).thenReturn(channel);
        when(clientEnvironment.getOwnerId()).thenReturn(new OwnerId(OWNER_ID));

        accountingClient = new AccountingClientImpl(clientEnvironment);

        Field field = ReflectionUtils.findFields(AccountingClientImpl.class,
                        f -> f.getName().equals("serviceBlockingStub"),
                        ReflectionUtils.HierarchyTraversalMode.TOP_DOWN)
                .get(0);

        field.setAccessible(true);
        field.set(accountingClient, AccountingServiceClient);

        verifierMock = Mockito.mockStatic(Verifier.class);
        requestConstructorMock = Mockito.mockStatic(RequestConstructor.class);
        requestSignerMock = Mockito.mockStatic(RequestSigner.class);
    }

    @AfterEach
    void cleanUp() {
        verifierMock.close();
        requestConstructorMock.close();
        requestSignerMock.close();
    }

    @Test
    void getBalance_success() {
        //Given
        var response = AccountingGenerator.generateBalanceResponse();

        var captor = ArgumentCaptor.forClass(Service.BalanceRequest.class);

        when(AccountingServiceClient.balance(captor.capture())).thenReturn(response);

        //When
        var result = accountingClient.getBalance(new CallContext(0, null));

        //Then
        requestConstructorMock.verify(
                () -> RequestConstructor.addMetaHeader(any(Service.BalanceRequest.Builder.class)),
                times(1)
        );
        requestSignerMock.verify(
                () -> RequestSigner.sign(any(Service.BalanceRequest.Builder.class), eq(null)),
                times(1)
        );
        verifierMock.verify(() -> Verifier.checkResponse(response), times(1));

        assertEquals(response.getBody().getBalance(), result);

        var request = captor.getValue();
        assertEquals(OWNER_ID, Base58.encode(request.getBody().getOwnerId().getValue().toByteArray()));
    }
}
