package info.frostfs.sdk.services;

import frostfs.apemanager.APEManagerServiceGrpc;
import frostfs.apemanager.Service;
import info.frostfs.sdk.dto.ape.*;
import info.frostfs.sdk.dto.chain.ChainTarget;
import info.frostfs.sdk.enums.*;
import info.frostfs.sdk.exceptions.ValidationFrostFSException;
import info.frostfs.sdk.jdo.ClientEnvironment;
import info.frostfs.sdk.jdo.parameters.CallContext;
import info.frostfs.sdk.jdo.parameters.ape.PrmApeChainAdd;
import info.frostfs.sdk.jdo.parameters.ape.PrmApeChainList;
import info.frostfs.sdk.jdo.parameters.ape.PrmApeChainRemove;
import info.frostfs.sdk.services.impl.ApeManagerClientImpl;
import info.frostfs.sdk.testgenerator.ApeManagerGenerator;
import info.frostfs.sdk.tools.RequestConstructor;
import info.frostfs.sdk.tools.RequestSigner;
import info.frostfs.sdk.tools.Verifier;
import io.grpc.Channel;
import org.apache.commons.lang3.ArrayUtils;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.junit.platform.commons.util.ReflectionUtils;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.MockedStatic;
import org.mockito.Mockito;
import org.mockito.junit.jupiter.MockitoExtension;

import java.lang.reflect.Field;
import java.nio.charset.StandardCharsets;
import java.util.Base64;

import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.mockito.Mockito.*;

@ExtendWith(MockitoExtension.class)
class ApeManagerClientTest {
    private static final String CHAIN_BASE64 =
            "AAAaY2hhaW4taWQtdGVzdAIAAAISR2V0T2JqZWN0AAIebmF0aXZlOm9iamVjdC8qAAIAABREZXBhcnRtZW50BEhSAA==";

    private ApeManagerClientImpl apeManagerClient;

    @Mock
    private APEManagerServiceGrpc.APEManagerServiceBlockingStub apeManagerServiceClient;
    @Mock
    private ClientEnvironment clientEnvironment;
    @Mock
    private Channel channel;

    private MockedStatic<Verifier> verifierMock;
    private MockedStatic<RequestConstructor> requestConstructorMock;
    private MockedStatic<RequestSigner> requestSignerMock;

    @BeforeEach
    void setUp() throws IllegalAccessException {
        when(clientEnvironment.getChannel()).thenReturn(channel);
        apeManagerClient = new ApeManagerClientImpl(clientEnvironment);

        Field field = ReflectionUtils.findFields(ApeManagerClientImpl.class,
                        f -> f.getName().equals("apeManagerServiceClient"),
                        ReflectionUtils.HierarchyTraversalMode.TOP_DOWN)
                .get(0);

        field.setAccessible(true);
        field.set(apeManagerClient, apeManagerServiceClient);

        verifierMock = Mockito.mockStatic(Verifier.class);
        requestConstructorMock = Mockito.mockStatic(RequestConstructor.class);
        requestSignerMock = Mockito.mockStatic(RequestSigner.class);
    }

    @AfterEach
    void cleanUp() {
        verifierMock.close();
        requestConstructorMock.close();
        requestSignerMock.close();
    }

    @Test
    void addChain_success() {
        //Given
        Chain chain = generateChain();
        ChainTarget chainTarget = generateChainTarget();
        PrmApeChainAdd params = new PrmApeChainAdd(chain, chainTarget);

        var response = ApeManagerGenerator.generateAddChainResponse();

        var captor = ArgumentCaptor.forClass(Service.AddChainRequest.class);

        when(apeManagerServiceClient.addChain(captor.capture())).thenReturn(response);

        //When
        var result = apeManagerClient.addChain(params, new CallContext(0, null));

        //Then
        requestConstructorMock.verify(
                () -> RequestConstructor.addMetaHeader(any(Service.AddChainRequest.Builder.class), eq(null)),
                times(1)
        );
        requestSignerMock.verify(
                () -> RequestSigner.sign(any(Service.AddChainRequest.Builder.class), eq(null)),
                times(1)
        );
        verifierMock.verify(() -> Verifier.checkResponse(response), times(1));

        assertThat(result).containsOnly(response.getBody().getChainId().toByteArray());

        var request = captor.getValue();
        assertEquals(
                Base64.getEncoder().encodeToString(request.getBody().getChain().getRaw().toByteArray()), CHAIN_BASE64)
        ;
        assertEquals(chainTarget.getName(), request.getBody().getTarget().getName());
        assertEquals(chainTarget.getType().value, request.getBody().getTarget().getType().getNumber());
    }

    @Test
    void addChain_wrongParams() {
        //Given
        Chain chain = generateChain();
        ChainTarget chainTarget = generateChainTarget();
        PrmApeChainAdd params1 = new PrmApeChainAdd(null, chainTarget);
        PrmApeChainAdd params2 = new PrmApeChainAdd(chain, null);
        PrmApeChainAdd params3 = new PrmApeChainAdd(null, null);


        //When + Then
        assertThrows(ValidationFrostFSException.class, () -> apeManagerClient.addChain(params1, new CallContext()));
        assertThrows(ValidationFrostFSException.class, () -> apeManagerClient.addChain(params2, new CallContext()));
        assertThrows(ValidationFrostFSException.class, () -> apeManagerClient.addChain(params3, new CallContext()));
    }

    @Test
    void removeChain_success() {
        //Given
        Chain chain = generateChain();
        ChainTarget chainTarget = generateChainTarget();
        PrmApeChainRemove params = new PrmApeChainRemove(Base64.getDecoder().decode(CHAIN_BASE64), chainTarget);

        var response = ApeManagerGenerator.generateRemoveChainResponse();

        var captor = ArgumentCaptor.forClass(Service.RemoveChainRequest.class);

        when(apeManagerServiceClient.removeChain(captor.capture())).thenReturn(response);

        //When
        apeManagerClient.removeChain(params, new CallContext(0, null));

        //Then
        requestConstructorMock.verify(
                () -> RequestConstructor.addMetaHeader(any(Service.RemoveChainRequest.Builder.class), eq(null)),
                times(1)
        );
        requestSignerMock.verify(
                () -> RequestSigner.sign(any(Service.RemoveChainRequest.Builder.class), eq(null)),
                times(1)
        );
        verifierMock.verify(() -> Verifier.checkResponse(response), times(1));

        var request = captor.getValue();
        assertThat(request.getBody().getChainId().toByteArray()).containsOnly(Base64.getDecoder().decode(CHAIN_BASE64));
        assertEquals(chainTarget.getName(), request.getBody().getTarget().getName());
        assertEquals(chainTarget.getType().value, request.getBody().getTarget().getType().getNumber());
    }

    @Test
    void removeChain_wrongParams() {
        //Given
        Chain chain = generateChain();
        ChainTarget chainTarget = generateChainTarget();
        PrmApeChainRemove params1 = new PrmApeChainRemove(null, chainTarget);
        PrmApeChainRemove params2 = new PrmApeChainRemove(Base64.getDecoder().decode(CHAIN_BASE64), null);
        PrmApeChainRemove params3 = new PrmApeChainRemove(null, null);

        //When + Then
        assertThrows(ValidationFrostFSException.class, () -> apeManagerClient.removeChain(params1, new CallContext()));
        assertThrows(ValidationFrostFSException.class, () -> apeManagerClient.removeChain(params2, new CallContext()));
        assertThrows(ValidationFrostFSException.class, () -> apeManagerClient.removeChain(params3, new CallContext()));
    }

    @Test
    void listChain_success() {
        //Given
        ChainTarget chainTarget = generateChainTarget();
        PrmApeChainList params = new PrmApeChainList(chainTarget);

        var response = ApeManagerGenerator.generateListChainsResponse();

        var captor = ArgumentCaptor.forClass(Service.ListChainsRequest.class);

        when(apeManagerServiceClient.listChains(captor.capture())).thenReturn(response);

        //When
        var result = apeManagerClient.listChains(params, new CallContext(0, null));

        //Then
        requestConstructorMock.verify(
                () -> RequestConstructor.addMetaHeader(any(Service.ListChainsRequest.Builder.class), eq(null)),
                times(1)
        );
        requestSignerMock.verify(
                () -> RequestSigner.sign(any(Service.ListChainsRequest.Builder.class), eq(null)),
                times(1)
        );
        verifierMock.verify(() -> Verifier.checkResponse(response), times(1));

        assertThat(result).hasSize(10);

        var request = captor.getValue();
        assertEquals(chainTarget.getName(), request.getBody().getTarget().getName());
        assertEquals(chainTarget.getType().value, request.getBody().getTarget().getType().getNumber());
    }

    @Test
    void listChain_wrongParams() {
        //When + Then
        assertThrows(ValidationFrostFSException.class,
                () -> apeManagerClient.listChains(new PrmApeChainList(null), new CallContext()));
    }

    private Chain generateChain() {
        var resources = new Resources(false, new String[]{"native:object/*"});
        var actions = new Actions(false, new String[]{"GetObject"});
        var condition = new Condition(
                ConditionType.COND_STRING_EQUALS, ConditionKindType.RESOURCE, "Department", "HR"
        );

        var rule = new Rule();
        rule.setStatus(RuleStatus.ALLOW);
        rule.setResources(resources);
        rule.setActions(actions);
        rule.setAny(false);
        rule.setConditions(new Condition[]{condition});

        var chain = new Chain();
        chain.setId(ArrayUtils.toObject("chain-id-test".getBytes(StandardCharsets.UTF_8)));
        chain.setRules(new Rule[]{rule});
        chain.setMatchType(RuleMatchType.DENY_PRIORITY);
        return chain;
    }

    private ChainTarget generateChainTarget() {
        return new ChainTarget("BzQw5HH3feoxFDD5tCT87Y1726qzgLfxEE7wgtoRzB3R", TargetType.NAMESPACE);
    }
}
