package info.frostfs.sdk.tools.ape;

import info.frostfs.sdk.dto.ape.*;
import info.frostfs.sdk.enums.ConditionKindType;
import info.frostfs.sdk.enums.ConditionType;
import info.frostfs.sdk.enums.RuleMatchType;
import info.frostfs.sdk.enums.RuleStatus;
import info.frostfs.sdk.exceptions.FrostFSException;
import info.frostfs.sdk.exceptions.ValidationFrostFSException;
import org.apache.commons.lang3.ArrayUtils;
import org.bouncycastle.util.encoders.Base64;
import org.junit.jupiter.api.Test;

import java.nio.charset.StandardCharsets;

import static java.util.Objects.isNull;
import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.*;

public class ApeRuleTest {

    @Test
    void apeRuleTest() {
        //Given
        var resources = new Resources(false, new String[]{"native:object/*"});
        var actions = new Actions(false, new String[]{"*"});
        var rule = new Rule();
        rule.setStatus(RuleStatus.ALLOW);
        rule.setResources(resources);
        rule.setActions(actions);
        rule.setAny(false);
        rule.setConditions(new Condition[]{});


        var chain = new Chain();
        chain.setId(ArrayUtils.toObject("chain-id-test".getBytes(StandardCharsets.UTF_8)));
        chain.setRules(new Rule[]{rule});
        chain.setMatchType(RuleMatchType.DENY_PRIORITY);

        //When
        var serialized = RuleSerializer.serialize(chain);
        var t = Base64.encode(serialized);
        var restoredChain = RuleDeserializer.deserialize(serialized);

        //Then
        assertThat(restoredChain.getId()).isNotEmpty().containsOnly(chain.getId());
        assertEquals(chain.getMatchType(), restoredChain.getMatchType());
        compareRules(chain.getRules(), restoredChain.getRules());
    }

    @Test
    void apeRuleTest2() {
        //Given
        var resources = new Resources(true, new String[]{"native:object/*", "U.S.S. ENTERPRISE"});
        var actions = new Actions(true, new String[]{"put", "get"});
        var cond1 = new Condition(
                ConditionType.COND_STRING_EQUALS, ConditionKindType.RESOURCE, "key1", "value1"
        );
        var cond2 = new Condition(
                ConditionType.COND_NUMERIC_GREATER_THAN, ConditionKindType.REQUEST, "key2", "value2"
        );
        var rule = new Rule();
        rule.setStatus(RuleStatus.ACCESS_DENIED);
        rule.setResources(resources);
        rule.setActions(actions);
        rule.setAny(true);
        rule.setConditions(new Condition[]{cond1, cond2});


        var chain = new Chain();
        chain.setId(ArrayUtils.toObject("dumptext".getBytes(StandardCharsets.UTF_8)));
        chain.setRules(new Rule[]{rule});
        chain.setMatchType(RuleMatchType.FIRST_MATCH);

        //When
        var serialized = RuleSerializer.serialize(chain);
        var restoredChain = RuleDeserializer.deserialize(serialized);

        //Then
        assertThat(restoredChain.getId()).isNotEmpty().containsOnly(chain.getId());
        assertEquals(chain.getMatchType(), restoredChain.getMatchType());
        compareRules(chain.getRules(), restoredChain.getRules());
    }

    @Test
    void apeRuleTest3() {
        //Given
        var chain = new Chain();
        chain.setMatchType(RuleMatchType.DENY_PRIORITY);

        //When
        var serialized = RuleSerializer.serialize(chain);
        var restoredChain = RuleDeserializer.deserialize(serialized);

        //Then
        assertNull(restoredChain.getId());
        assertEquals(chain.getMatchType(), restoredChain.getMatchType());
        assertNull(restoredChain.getRules());
    }

    @Test
    void apeRule_deserialize_wrong() {
        //When + Then
        assertThrows(ValidationFrostFSException.class, () -> RuleDeserializer.deserialize(null));
        assertThrows(ValidationFrostFSException.class, () -> RuleDeserializer.deserialize(new byte[]{}));
        assertThrows(FrostFSException.class, () -> RuleDeserializer.deserialize(new byte[]{1, 2, 3}));
        assertThrows(ValidationFrostFSException.class, () -> RuleDeserializer.deserialize(new byte[]{
                0x00, 0x00, 0x3A, 0x77, 0x73, 0x3A, 0x69, 0x61, 0x6D, 0x3A, 0x3A, 0x6E, 0x61, 0x6D, 0x65, 0x73,
                0x70, 0x61, 0x63, 0x65, 0x3A, 0x67, 0x72, 0x6F, 0x75, 0x70, 0x2F, 0x73, 0x6F, (byte) 0x82, (byte) 0x82,
                (byte) 0x82, (byte) 0x82, (byte) 0x82, (byte) 0x82, 0x75, (byte) 0x82
        }));
    }

    @Test
    void apeRule_serialize_wrong() {
        //When + Then
        assertThrows(ValidationFrostFSException.class, () -> RuleSerializer.serialize(null));
    }

    private void compareRules(Rule[] rules1, Rule[] rules2) {
        assertThat(rules1).isNotEmpty();
        assertThat(rules2).isNotEmpty();

        assertEquals(rules1.length, rules2.length);

        for (int ri = 0; ri < rules1.length; ri++) {
            var rule1 = rules1[ri];
            var rule2 = rules2[ri];

            assertEquals(rule1.getStatus(), rule2.getStatus());
            assertEquals(rule1.isAny(), rule2.isAny());

            compareActions(rule1.getActions(), rule2.getActions());
            compareResources(rule1.getResources(), rule2.getResources());
            compareConditions(rule1.getConditions(), rule2.getConditions());
        }
    }

    private void compareActions(Actions actions1, Actions actions2) {
        if (isNull(actions1) && isNull(actions2)) {
            return;
        }

        assertEquals(actions1.isInverted(), actions2.isInverted());
        if (ArrayUtils.isEmpty(actions1.getNames()) && ArrayUtils.isEmpty(actions2.getNames())) {
            return;
        }

        assertThat(actions2.getNames()).hasSize(actions1.getNames().length).containsOnly(actions1.getNames());
    }

    private void compareResources(Resources resources1, Resources resources2) {
        if (isNull(resources1) && isNull(resources2)) {
            return;
        }

        assertEquals(resources1.isInverted(), resources2.isInverted());
        if (ArrayUtils.isEmpty(resources1.getNames()) && ArrayUtils.isEmpty(resources2.getNames())) {
            return;
        }

        assertThat(resources2.getNames()).hasSize(resources1.getNames().length).containsOnly(resources1.getNames());
    }

    private void compareConditions(Condition[] conditions1, Condition[] conditions2) {
        if (ArrayUtils.isEmpty(conditions1) && ArrayUtils.isEmpty(conditions2)) {
            return;
        }

        assertEquals(conditions1.length, conditions2.length);
        for (int i = 0; i < conditions1.length; i++) {
            assertEquals(conditions1[i].getOp(), conditions2[i].getOp());
            assertEquals(conditions1[i].getKind(), conditions2[i].getKind());
            assertEquals(conditions1[i].getKey(), conditions2[i].getKey());
            assertEquals(conditions1[i].getValue(), conditions2[i].getValue());
        }
    }

}
