package info.frostfs.sdk.mappers.netmap;

import frostfs.netmap.Types;
import info.frostfs.sdk.dto.netmap.Filter;
import info.frostfs.sdk.enums.FilterOperation;
import info.frostfs.sdk.exceptions.ProcessFrostFSException;
import info.frostfs.sdk.exceptions.ValidationFrostFSException;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.ArrayUtils;

import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;

import static info.frostfs.sdk.constants.ErrorConst.INPUT_PARAM_IS_MISSING_TEMPLATE;
import static info.frostfs.sdk.constants.ErrorConst.UNKNOWN_ENUM_VALUE_TEMPLATE;
import static java.util.Objects.isNull;

public class FilterMapper {
    private FilterMapper() {
    }

    public static List<Types.Filter> toGrpcMessages(Filter[] filters) {
        if (ArrayUtils.isEmpty(filters)) {
            return Collections.emptyList();
        }

        return Arrays.stream(filters).map(FilterMapper::toGrpcMessage).collect(Collectors.toList());
    }

    public static Types.Filter toGrpcMessage(Filter filter) {
        if (isNull(filter)) {
            throw new ValidationFrostFSException(
                    String.format(INPUT_PARAM_IS_MISSING_TEMPLATE, Filter.class.getName())
            );
        }

        var operation = Types.Operation.forNumber(filter.getOperation().value);
        if (isNull(operation)) {
            throw new ProcessFrostFSException(String.format(
                    UNKNOWN_ENUM_VALUE_TEMPLATE,
                    Types.Operation.class.getName(),
                    filter.getOperation().name()
            ));
        }

        return Types.Filter.newBuilder()
                .setName(filter.getName())
                .setKey(filter.getKey())
                .setOp(operation)
                .setValue(filter.getValue())
                .addAllFilters(toGrpcMessages(filter.getFilters()))
                .build();
    }

    public static Filter[] toModels(List<Types.Filter> filters) {
        if (CollectionUtils.isEmpty(filters)) {
            return null;
        }

        return filters.stream().map(FilterMapper::toModel).toArray(Filter[]::new);
    }

    public static Filter toModel(Types.Filter filter) {
        if (isNull(filter) || filter.getSerializedSize() == 0) {
            return null;
        }

        var operation = FilterOperation.get(filter.getOpValue());
        if (isNull(operation)) {
            throw new ProcessFrostFSException(
                    String.format(UNKNOWN_ENUM_VALUE_TEMPLATE, FilterOperation.class.getName(), filter.getOp())
            );
        }

        return new Filter(
                filter.getName(),
                filter.getKey(),
                operation,
                filter.getValue(),
                toModels(filter.getFiltersList())
        );
    }
}
