package info.frostfs.sdk.mappers.netmap;

import frostfs.netmap.Types;
import info.frostfs.sdk.dto.netmap.Selector;
import info.frostfs.sdk.enums.SelectorClause;
import info.frostfs.sdk.exceptions.ProcessFrostFSException;
import info.frostfs.sdk.exceptions.ValidationFrostFSException;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.ArrayUtils;

import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;

import static info.frostfs.sdk.constants.ErrorConst.INPUT_PARAM_IS_MISSING_TEMPLATE;
import static info.frostfs.sdk.constants.ErrorConst.UNKNOWN_ENUM_VALUE_TEMPLATE;
import static java.util.Objects.isNull;

public class SelectorMapper {
    private SelectorMapper() {
    }

    public static List<Types.Selector> toGrpcMessages(Selector[] selectors) {
        if (ArrayUtils.isEmpty(selectors)) {
            return Collections.emptyList();
        }

        return Arrays.stream(selectors).map(SelectorMapper::toGrpcMessage).collect(Collectors.toList());
    }

    public static Types.Selector toGrpcMessage(Selector selector) {
        if (isNull(selector)) {
            throw new ValidationFrostFSException(
                    String.format(INPUT_PARAM_IS_MISSING_TEMPLATE, Selector.class.getName())
            );
        }

        var clause = Types.Clause.forNumber(selector.getClause().value);
        if (isNull(clause)) {
            throw new ProcessFrostFSException(String.format(
                    UNKNOWN_ENUM_VALUE_TEMPLATE,
                    Types.Clause.class.getName(),
                    selector.getClause().name()
            ));
        }

        return Types.Selector.newBuilder()
                .setName(selector.getName())
                .setCount(selector.getCount())
                .setClause(clause)
                .setAttribute(selector.getAttribute())
                .setFilter(selector.getFilter())
                .build();
    }

    public static Selector[] toModels(List<Types.Selector> selectors) {
        if (CollectionUtils.isEmpty(selectors)) {
            return null;
        }

        return selectors.stream().map(SelectorMapper::toModel).toArray(Selector[]::new);
    }

    public static Selector toModel(Types.Selector selector) {
        if (isNull(selector) || selector.getSerializedSize() == 0) {
            return null;
        }

        var clause = SelectorClause.get(selector.getClauseValue());
        if (isNull(clause)) {
            throw new ProcessFrostFSException(
                    String.format(UNKNOWN_ENUM_VALUE_TEMPLATE, SelectorClause.class.getName(), selector.getClause())
            );
        }

        return new Selector(
                selector.getName(),
                selector.getCount(),
                clause,
                selector.getAttribute(),
                selector.getFilter()
        );
    }
}
