package info.frostfs.sdk.mappers.object;

import com.google.protobuf.ByteString;
import frostfs.object.Types;
import info.frostfs.sdk.dto.container.ContainerId;
import info.frostfs.sdk.dto.object.ObjectAttribute;
import info.frostfs.sdk.dto.object.ObjectHeader;
import info.frostfs.sdk.enums.ObjectType;
import info.frostfs.sdk.exceptions.ProcessFrostFSException;
import info.frostfs.sdk.exceptions.ValidationFrostFSException;
import info.frostfs.sdk.mappers.container.ContainerIdMapper;
import info.frostfs.sdk.mappers.netmap.VersionMapper;
import org.apache.commons.collections4.ListUtils;

import java.util.stream.Collectors;

import static info.frostfs.sdk.constants.ErrorConst.INPUT_PARAM_IS_MISSING_TEMPLATE;
import static info.frostfs.sdk.constants.ErrorConst.UNKNOWN_ENUM_VALUE_TEMPLATE;
import static java.util.Objects.isNull;
import static java.util.Objects.nonNull;

public class ObjectHeaderMapper {
    private ObjectHeaderMapper() {
    }

    public static Types.Header toGrpcMessage(ObjectHeader header) {
        if (isNull(header)) {
            throw new ValidationFrostFSException(
                    String.format(INPUT_PARAM_IS_MISSING_TEMPLATE, ObjectHeader.class.getName())
            );
        }

        var objectType = Types.ObjectType.forNumber(header.getObjectType().value);
        if (isNull(objectType)) {
            throw new ProcessFrostFSException(String.format(
                    UNKNOWN_ENUM_VALUE_TEMPLATE,
                    Types.ObjectType.class.getName(),
                    header.getObjectType().name()
            ));
        }

        var head = Types.Header.newBuilder()
                .setOwnerId(OwnerIdMapper.toGrpcMessage(header.getOwnerId()))
                .setVersion(VersionMapper.toGrpcMessage(header.getVersion()))
                .setContainerId(ContainerIdMapper.toGrpcMessage(header.getContainerId()))
                .setObjectType(objectType)
                .setPayloadLength(header.getPayloadLength());

        for (ObjectAttribute objectAttribute : header.getAttributes()) {
            head.addAttributes(ObjectAttributeMapper.toGrpcMessage(objectAttribute));
        }

        if (nonNull(header.getSplit())) {
            var grpcSplit = Types.Header.Split.newBuilder()
                    .setSplitId(ByteString.copyFrom(header.getSplit().getSplitId().toBinary()))
                    .build();
            head.setSplit(grpcSplit);
        }

        return head.build();
    }

    public static ObjectHeader toModel(Types.Header header) {
        if (isNull(header) || header.getSerializedSize() == 0) {
            return null;
        }

        var objectType = ObjectType.get(header.getObjectTypeValue());
        if (isNull(objectType)) {
            throw new ProcessFrostFSException(
                    String.format(UNKNOWN_ENUM_VALUE_TEMPLATE, ObjectType.class.getName(), header.getObjectType())
            );
        }

        return new ObjectHeader(
                new ContainerId(header.getContainerId().getValue().toByteArray()),
                objectType,
                ListUtils.emptyIfNull(header.getAttributesList()).stream()
                        .map(ObjectAttributeMapper::toModel)
                        .collect(Collectors.toList()),
                header.getPayloadLength(),
                VersionMapper.toModel(header.getVersion())
        );
    }
}
