package info.frostfs.sdk.mappers.container;

import com.google.protobuf.ByteString;
import frostfs.container.Types;
import info.frostfs.sdk.dto.container.Container;
import info.frostfs.sdk.dto.netmap.PlacementPolicy;
import info.frostfs.sdk.dto.netmap.Replica;
import info.frostfs.sdk.dto.netmap.Version;
import info.frostfs.sdk.dto.object.OwnerId;
import info.frostfs.sdk.mappers.object.OwnerIdMapper;
import org.junit.jupiter.api.Test;

import java.util.UUID;

import static info.frostfs.sdk.UuidExtension.asBytes;
import static info.frostfs.sdk.UuidExtension.asUuid;
import static info.frostfs.sdk.constants.AttributeConst.DISABLE_HOMOMORPHIC_HASHING_ATTRIBUTE;
import static org.junit.jupiter.api.Assertions.*;

public class ContainerMapperTest {
    private static final String OWNER_ID = "NVxUSpEEJzYXZZtUs18PrJTD9QZkLLNQ8S";

    @Test
    void toGrpcMessage_successFullMessage() {
        //Given
        var placementPolicy = new PlacementPolicy(new Replica[]{new Replica(3)}, true, 1);
        var container = new Container(placementPolicy);
        container.getAttributes().put("key1", "val1");
        container.getAttributes().put(DISABLE_HOMOMORPHIC_HASHING_ATTRIBUTE, "false");
        container.setVersion(new Version());
        container.setOwnerId(new OwnerId(OWNER_ID));

        //When
        var result = ContainerMapper.toGrpcMessage(container);

        //Then
        assertNotNull(result);
        assertEquals(container.getNonce(), asUuid(result.getNonce().toByteArray()));
        assertEquals(container.getPlacementPolicy().isUnique(), result.getPlacementPolicy().getUnique());
        assertEquals(
                container.getPlacementPolicy().getBackupFactory(),
                result.getPlacementPolicy().getContainerBackupFactor()
        );
        assertEquals(placementPolicy.getReplicas().length, result.getPlacementPolicy().getReplicasCount());
        assertEquals(
                container.getPlacementPolicy().getReplicas()[0].getCount(),
                result.getPlacementPolicy().getReplicasList().get(0).getCount()
        );
        assertEquals(
                container.getPlacementPolicy().getReplicas()[0].getSelector(),
                result.getPlacementPolicy().getReplicasList().get(0).getSelector()
        );

        assertEquals("key1", result.getAttributes(0).getKey());
        assertEquals("val1", result.getAttributes(0).getValue());
        assertEquals(DISABLE_HOMOMORPHIC_HASHING_ATTRIBUTE, result.getAttributes(1).getKey());
        assertEquals("false", result.getAttributes(1).getValue());

        assertEquals(container.getVersion().getMajor(), result.getVersion().getMajor());
        assertEquals(container.getVersion().getMinor(), result.getVersion().getMinor());
        assertEquals(container.getOwnerId().toString(), OwnerIdMapper.toModel(result.getOwnerId()).toString());
    }

    @Test
    void toGrpcMessage_success() {
        //Given
        var placementPolicy = new PlacementPolicy(new Replica[]{new Replica(3)}, true, 1);
        var container = new Container(placementPolicy);
        container.getAttributes().put("key1", "val1");
        container.getAttributes().put(DISABLE_HOMOMORPHIC_HASHING_ATTRIBUTE, "false");

        //When
        var result = ContainerMapper.toGrpcMessage(container);

        //Then
        assertNotNull(result);
        assertEquals(container.getNonce(), asUuid(result.getNonce().toByteArray()));
        assertEquals(container.getPlacementPolicy().isUnique(), result.getPlacementPolicy().getUnique());
        assertEquals(
                container.getPlacementPolicy().getBackupFactory(),
                result.getPlacementPolicy().getContainerBackupFactor()
        );
        assertEquals(placementPolicy.getReplicas().length, result.getPlacementPolicy().getReplicasCount());
        assertEquals(
                container.getPlacementPolicy().getReplicas()[0].getCount(),
                result.getPlacementPolicy().getReplicasList().get(0).getCount()
        );
        assertEquals(
                container.getPlacementPolicy().getReplicas()[0].getSelector(),
                result.getPlacementPolicy().getReplicasList().get(0).getSelector()
        );

        assertEquals("key1", result.getAttributes(0).getKey());
        assertEquals("val1", result.getAttributes(0).getValue());
        assertEquals(DISABLE_HOMOMORPHIC_HASHING_ATTRIBUTE, result.getAttributes(1).getKey());
        assertEquals("false", result.getAttributes(1).getValue());
    }

    @Test
    void toGrpcMessage_null() {
        //When + Then
        assertNull(ContainerMapper.toGrpcMessage(null));
    }

    @Test
    void toModel_success() {
        //Given
        var version = frostfs.refs.Types.Version.newBuilder()
                .setMajor(1)
                .setMinor(2)
                .build();

        var replica = frostfs.netmap.Types.Replica.newBuilder()
                .setCount(1)
                .setSelector("")
                .build();

        var placementPolicy = frostfs.netmap.Types.PlacementPolicy.newBuilder()
                .setUnique(true)
                .addReplicas(replica)
                .setContainerBackupFactor(2)
                .build();

        var attribute1 = Types.Container.Attribute.newBuilder()
                .setKey("key1")
                .setValue("val1")
                .build();

        var attribute2 = Types.Container.Attribute.newBuilder()
                .setKey("key2")
                .setValue("val2")
                .build();

        var container = Types.Container.newBuilder()
                .setNonce(ByteString.copyFrom(asBytes(UUID.randomUUID())))
                .setVersion(version)
                .setPlacementPolicy(placementPolicy)
                .addAttributes(attribute1)
                .addAttributes(attribute2)
                .build();

        //When
        var result = ContainerMapper.toModel(container);

        //Then
        assertNotNull(result);
        assertEquals(asUuid(container.getNonce().toByteArray()), result.getNonce());
        assertEquals(container.getPlacementPolicy().getUnique(), result.getPlacementPolicy().isUnique());
        assertEquals(
                container.getPlacementPolicy().getContainerBackupFactor(),
                result.getPlacementPolicy().getBackupFactory()
        );
        assertEquals(placementPolicy.getReplicasCount(), result.getPlacementPolicy().getReplicas().length);
        assertEquals(
                container.getPlacementPolicy().getReplicasList().get(0).getCount(),
                result.getPlacementPolicy().getReplicas()[0].getCount()
        );
        assertEquals(
                container.getPlacementPolicy().getReplicasList().get(0).getSelector(),
                result.getPlacementPolicy().getReplicas()[0].getSelector()
        );
        assertEquals(version.getMajor(), result.getVersion().getMajor());
        assertEquals(version.getMinor(), result.getVersion().getMinor());

        assertEquals(attribute1.getValue(), result.getAttributes().get(attribute1.getKey()));
        assertEquals(attribute2.getValue(), result.getAttributes().get(attribute2.getKey()));
    }

    @Test
    void toModel_null() {
        //When + Then
        assertNull(ContainerMapper.toModel(null));
        assertNull(ContainerMapper.toModel(Types.Container.getDefaultInstance()));
    }
}
