package info.frostfs.sdk.mappers.netmap;

import frostfs.netmap.Types;
import info.frostfs.sdk.dto.netmap.Filter;
import info.frostfs.sdk.enums.FilterOperation;
import info.frostfs.sdk.exceptions.ProcessFrostFSException;
import info.frostfs.sdk.exceptions.ValidationFrostFSException;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.mockito.MockedStatic;

import java.util.Collections;
import java.util.List;

import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.mockStatic;

public class FilterMapperTest {

    @ParameterizedTest
    @EnumSource(value = FilterOperation.class)
    void toGrpcMessages_success(FilterOperation operation) {
        //Given
        var filter1 = new Filter("name1", "key1", operation, "value1", null);
        var filter2 = new Filter("name2", "key2", operation, "value2", null);

        //When
        var result = FilterMapper.toGrpcMessages(new Filter[]{filter1, filter2});

        //Then
        assertThat(result).isNotNull().hasSize(2);

        assertEquals(filter1.getKey(), result.get(0).getKey());
        assertEquals(filter1.getName(), result.get(0).getName());
        assertEquals(filter1.getOperation().value, result.get(0).getOpValue());
        assertEquals(filter1.getValue(), result.get(0).getValue());
        assertEquals(0, result.get(0).getFiltersCount());

        assertEquals(filter2.getKey(), result.get(1).getKey());
        assertEquals(filter2.getName(), result.get(1).getName());
        assertEquals(filter2.getOperation().value, result.get(1).getOpValue());
        assertEquals(filter2.getValue(), result.get(1).getValue());
        assertEquals(0, result.get(1).getFiltersCount());
    }

    @Test
    void toGrpcMessages_null() {
        //When + Then
        assertEquals(Collections.emptyList(), FilterMapper.toGrpcMessages(null));
        assertEquals(Collections.emptyList(), FilterMapper.toGrpcMessages(new Filter[]{}));
    }

    @ParameterizedTest
    @EnumSource(value = FilterOperation.class)
    void toGrpcMessage_success(FilterOperation operation) {
        //Given
        var filterChild = new Filter("name1", "key1", operation, "value1", null);
        var filterParent = new Filter("name2", "key2", operation, "value2", new Filter[]{filterChild});

        //When
        var result = FilterMapper.toGrpcMessage(filterParent);

        //Then
        assertNotNull(result);
        assertEquals(filterParent.getKey(), result.getKey());
        assertEquals(filterParent.getName(), result.getName());
        assertEquals(filterParent.getOperation().value, result.getOpValue());
        assertEquals(filterParent.getValue(), result.getValue());
        assertEquals(filterParent.getFilters().length, result.getFiltersCount());

        var filterChildGrpc = result.getFilters(0);
        assertEquals(filterChild.getKey(), filterChildGrpc.getKey());
        assertEquals(filterChild.getName(), filterChildGrpc.getName());
        assertEquals(filterChild.getOperation().value, filterChildGrpc.getOpValue());
        assertEquals(filterChild.getValue(), filterChildGrpc.getValue());
        assertEquals(0, filterChildGrpc.getFiltersCount());
    }

    @Test
    void toGrpcMessage_null() {
        //When + Then
        assertThrows(ValidationFrostFSException.class, () -> FilterMapper.toGrpcMessage(null));
    }

    @Test
    void toGrpcMessage_notValidOperation() {
        //Given
        var filter = new Filter("name1", "key1", FilterOperation.EQ, "value1", null);


        //When + Then
        try (MockedStatic<Types.Operation> mockStatic = mockStatic(Types.Operation.class)) {
            mockStatic.when(() -> Types.Operation.forNumber(filter.getOperation().value))
                    .thenReturn(null);

            assertThrows(ProcessFrostFSException.class, () -> FilterMapper.toGrpcMessage(filter));
        }
    }

    @ParameterizedTest
    @EnumSource(value = Types.Operation.class, names = "UNRECOGNIZED", mode = EnumSource.Mode.EXCLUDE)
    void toModels_success(Types.Operation operation) {
        //Given
        var filter1 = Types.Filter.newBuilder()
                .setName("name1")
                .setKey("key1")
                .setOp(operation)
                .setValue("value1")
                .build();
        var filter2 = Types.Filter.newBuilder()
                .setName("name2")
                .setKey("key2")
                .setOp(operation)
                .setValue("value2")
                .build();

        //When
        var result = FilterMapper.toModels(List.of(filter1, filter2));

        //Then
        assertThat(result).isNotNull().hasSize(2);


        assertNotNull(result);
        assertEquals(filter1.getKey(), result[0].getKey());
        assertEquals(filter1.getName(), result[0].getName());
        assertEquals(filter1.getOpValue(), result[0].getOperation().value);
        assertEquals(filter1.getValue(), result[0].getValue());
        assertNull(result[0].getFilters());

        assertEquals(filter2.getKey(), result[1].getKey());
        assertEquals(filter2.getName(), result[1].getName());
        assertEquals(filter2.getOpValue(), result[1].getOperation().value);
        assertEquals(filter2.getValue(), result[1].getValue());
        assertNull(result[1].getFilters());
    }

    @Test
    void toModels_null() {
        //When + Then
        assertNull(FilterMapper.toModels(null));
        assertNull(FilterMapper.toModels(Collections.emptyList()));
    }

    @ParameterizedTest
    @EnumSource(value = Types.Operation.class, names = "UNRECOGNIZED", mode = EnumSource.Mode.EXCLUDE)
    void toModel_success(Types.Operation operation) {
        //Given
        var filterChild = Types.Filter.newBuilder()
                .setName("name1")
                .setKey("key1")
                .setOp(operation)
                .setValue("value1")
                .build();
        var filterParent = Types.Filter.newBuilder()
                .setName("name2")
                .setKey("key2")
                .setOp(operation)
                .setValue("value2")
                .addFilters(filterChild)
                .build();

        //When
        var result = FilterMapper.toModel(filterParent);

        //Then
        assertNotNull(result);
        assertEquals(filterParent.getKey(), result.getKey());
        assertEquals(filterParent.getName(), result.getName());
        assertEquals(filterParent.getOpValue(), result.getOperation().value);
        assertEquals(filterParent.getValue(), result.getValue());
        assertEquals(filterParent.getFiltersCount(), result.getFilters().length);

        var filterChildModel = result.getFilters()[0];
        assertEquals(filterChild.getKey(), filterChildModel.getKey());
        assertEquals(filterChild.getName(), filterChildModel.getName());
        assertEquals(filterChild.getOpValue(), filterChildModel.getOperation().value);
        assertEquals(filterChild.getValue(), filterChildModel.getValue());
        assertNull(filterChildModel.getFilters());
    }

    @Test
    void toModel_null() {
        //When + Then
        assertNull(FilterMapper.toModel(null));
        assertNull(FilterMapper.toModel(Types.Filter.getDefaultInstance()));
    }

    @Test
    void toModel_notValidScheme() {
        //Given
        var filter = Types.Filter.newBuilder()
                .setName("name1")
                .setKey("key1")
                .setOp(Types.Operation.EQ)
                .setValue("value1")
                .build();

        //When + Then
        try (MockedStatic<FilterOperation> mockStatic = mockStatic(FilterOperation.class)) {
            mockStatic.when(() -> FilterOperation.get(Types.Operation.EQ.getNumber()))
                    .thenReturn(null);

            assertThrows(ProcessFrostFSException.class, () -> FilterMapper.toModel(filter));
        }
    }
}
