package info.frostfs.sdk.mappers.netmap;

import com.google.protobuf.ByteString;
import frostfs.netmap.Service;
import frostfs.netmap.Types;
import org.junit.jupiter.api.Test;

import java.util.Map;

import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.*;

public class NetmapSnapshotMapperTest {

    @Test
    void toModel_success() {
        //Given
        var version = frostfs.refs.Types.Version.newBuilder()
                .setMajor(1)
                .setMinor(2)
                .build();

        var nodeInfo1 = Types.NodeInfo.newBuilder()
                .setPublicKey(ByteString.copyFrom(new byte[]{1, 2, 3, 4, 5, 6}))
                .addAttributes(Types.NodeInfo.Attribute.newBuilder().setKey("key1").setValue("value1").build())
                .addAttributes(Types.NodeInfo.Attribute.newBuilder().setKey("key2").setValue("value2").build())
                .addAddresses("address1")
                .addAddresses("address2")
                .setState(Types.NodeInfo.State.ONLINE)
                .build();

        var nodeInfo2 = Types.NodeInfo.newBuilder()
                .setPublicKey(ByteString.copyFrom(new byte[]{7, 8, 9, 10}))
                .addAttributes(Types.NodeInfo.Attribute.newBuilder().setKey("key3").setValue("value3").build())
                .addAttributes(Types.NodeInfo.Attribute.newBuilder().setKey("key4").setValue("value4").build())
                .addAddresses("address3")
                .addAddresses("address4")
                .setState(Types.NodeInfo.State.MAINTENANCE)
                .build();

        var netmap = Types.Netmap.newBuilder()
                .setEpoch(10)
                .addNodes(nodeInfo1)
                .addNodes(nodeInfo2)
                .build();

        var body = Service.NetmapSnapshotResponse.Body.newBuilder()
                .setNetmap(netmap)
                .build();

        var response = Service.NetmapSnapshotResponse.newBuilder()
                .setBody(body)
                .setMetaHeader(frostfs.session.Types.ResponseMetaHeader.newBuilder().setVersion(version).build())
                .build();

        //When
        var result = NetmapSnapshotMapper.toModel(response);

        //Then
        assertNotNull(result);

        var resultNode1 = result.getNodeInfoCollection().get(0);
        var resultNode2 = result.getNodeInfoCollection().get(1);


        assertEquals(netmap.getEpoch(), result.getEpoch());
        assertEquals(netmap.getNodesCount(), result.getNodeInfoCollection().size());

        assertThat(resultNode1.getPublicKey()).containsExactly(nodeInfo1.getPublicKey().toByteArray());
        assertEquals(nodeInfo1.getState().getNumber(), resultNode1.getState().value);
        assertEquals(version.getMajor(), resultNode1.getVersion().getMajor());
        assertEquals(version.getMinor(), resultNode1.getVersion().getMinor());
        assertThat(resultNode1.getAddresses())
                .hasSize(2)
                .containsAll(nodeInfo1.getAddressesList());
        assertThat(resultNode1.getAttributes())
                .hasSize(2)
                .contains(Map.entry("key1", "value1"), Map.entry("key2", "value2"));

        assertThat(resultNode2.getPublicKey()).containsExactly(nodeInfo2.getPublicKey().toByteArray());
        assertEquals(nodeInfo2.getState().getNumber(), resultNode2.getState().value);
        assertEquals(version.getMajor(), resultNode2.getVersion().getMajor());
        assertEquals(version.getMinor(), resultNode2.getVersion().getMinor());
        assertThat(resultNode2.getAddresses())
                .hasSize(2)
                .containsAll(nodeInfo2.getAddressesList());
        assertThat(resultNode2.getAttributes())
                .hasSize(2)
                .contains(Map.entry("key3", "value3"), Map.entry("key4", "value4"));
    }

    @Test
    void toModel_null() {
        //When + Then
        assertNull(NetmapSnapshotMapper.toModel(null));
        assertNull(NetmapSnapshotMapper.toModel(Service.NetmapSnapshotResponse.getDefaultInstance()));
    }
}
