package info.frostfs.sdk.mappers.netmap;

import com.google.protobuf.ByteString;
import frostfs.netmap.Service;
import frostfs.netmap.Types;
import info.frostfs.sdk.enums.NodeState;
import info.frostfs.sdk.exceptions.ProcessFrostFSException;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.mockito.MockedStatic;

import java.util.Map;

import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.mockStatic;

public class NodeInfoMapperTest {

    @ParameterizedTest
    @EnumSource(value = Types.NodeInfo.State.class, names = "UNRECOGNIZED", mode = EnumSource.Mode.EXCLUDE)
    void toModel_success(Types.NodeInfo.State state) {
        //Given
        var version = frostfs.refs.Types.Version.newBuilder()
                .setMajor(1)
                .setMinor(2)
                .build();

        var nodeInfo = Types.NodeInfo.newBuilder()
                .setPublicKey(ByteString.copyFrom(new byte[]{1, 2, 3, 4, 5, 6}))
                .addAttributes(Types.NodeInfo.Attribute.newBuilder().setKey("key1").setValue("value1").build())
                .addAttributes(Types.NodeInfo.Attribute.newBuilder().setKey("key2").setValue("value2").build())
                .addAddresses("address1")
                .addAddresses("address2")
                .setState(state)
                .build();

        var body = Service.LocalNodeInfoResponse.Body.newBuilder()
                .setNodeInfo(nodeInfo)
                .setVersion(version)
                .build();

        //When
        var result = NodeInfoMapper.toModel(body);

        //Then
        assertNotNull(result);
        assertThat(result.getPublicKey()).containsExactly(nodeInfo.getPublicKey().toByteArray());
        assertEquals(nodeInfo.getState().getNumber(), result.getState().value);
        assertEquals(version.getMajor(), result.getVersion().getMajor());
        assertEquals(version.getMinor(), result.getVersion().getMinor());
        assertThat(result.getAddresses())
                .hasSize(2)
                .containsAll(nodeInfo.getAddressesList());
        assertThat(result.getAttributes())
                .hasSize(2)
                .contains(Map.entry("key1", "value1"), Map.entry("key2", "value2"));
    }

    @ParameterizedTest
    @EnumSource(value = Types.NodeInfo.State.class, names = "UNRECOGNIZED", mode = EnumSource.Mode.EXCLUDE)
    void toModel2_success(Types.NodeInfo.State state) {
        //Given
        var version = frostfs.refs.Types.Version.newBuilder()
                .setMajor(1)
                .setMinor(2)
                .build();

        var nodeInfo = Types.NodeInfo.newBuilder()
                .setPublicKey(ByteString.copyFrom(new byte[]{1, 2, 3, 4, 5, 6}))
                .addAttributes(Types.NodeInfo.Attribute.newBuilder().setKey("key1").setValue("value1").build())
                .addAttributes(Types.NodeInfo.Attribute.newBuilder().setKey("key2").setValue("value2").build())
                .addAddresses("address1")
                .addAddresses("address2")
                .setState(state)
                .build();

        //When
        var result = NodeInfoMapper.toModel(nodeInfo, version);

        //Then
        assertNotNull(result);
        assertThat(result.getPublicKey()).containsExactly(nodeInfo.getPublicKey().toByteArray());
        assertEquals(nodeInfo.getState().getNumber(), result.getState().value);
        assertEquals(version.getMajor(), result.getVersion().getMajor());
        assertEquals(version.getMinor(), result.getVersion().getMinor());
        assertThat(result.getAddresses())
                .hasSize(2)
                .containsAll(nodeInfo.getAddressesList());
        assertThat(result.getAttributes())
                .hasSize(2)
                .contains(Map.entry("key1", "value1"), Map.entry("key2", "value2"));
    }

    @Test
    void toModel_null() {
        //When + Then
        assertNull(NodeInfoMapper.toModel(null));
        assertNull(NodeInfoMapper.toModel(null, null));
        assertNull(NodeInfoMapper.toModel(Service.LocalNodeInfoResponse.Body.getDefaultInstance()));
        assertNull(NodeInfoMapper.toModel(Types.NodeInfo.getDefaultInstance(), null));
    }

    @Test
    void toModel_notValidScheme() {
        //Given
        var nodeInfo = Types.NodeInfo.newBuilder()
                .setState(Types.NodeInfo.State.ONLINE)
                .build();

        var body = Service.LocalNodeInfoResponse.Body.newBuilder()
                .setNodeInfo(nodeInfo)
                .build();

        //When + Then
        try (MockedStatic<NodeState> mockStatic = mockStatic(NodeState.class)) {
            mockStatic.when(() -> NodeState.get(Types.NodeInfo.State.ONLINE.getNumber()))
                    .thenReturn(null);

            assertThrows(ProcessFrostFSException.class, () -> NodeInfoMapper.toModel(body));
        }
    }
}
