package info.frostfs.sdk.mappers.netmap;

import frostfs.netmap.Types;
import info.frostfs.sdk.dto.netmap.Selector;
import info.frostfs.sdk.enums.SelectorClause;
import info.frostfs.sdk.exceptions.ProcessFrostFSException;
import info.frostfs.sdk.exceptions.ValidationFrostFSException;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.mockito.MockedStatic;

import java.util.Collections;
import java.util.List;

import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.mockStatic;

public class SelectorMapperTest {

    @ParameterizedTest
    @EnumSource(value = SelectorClause.class)
    void toGrpcMessages_success(SelectorClause clause) {
        //Given
        var selector1 = new Selector("name1", 1, clause, "attribute1", "filter1");
        var selector2 = new Selector("name2", 2, clause, "attribute2", "filter2");

        //When
        var result = SelectorMapper.toGrpcMessages(new Selector[]{selector1, selector2});

        //Then
        assertThat(result).isNotNull().hasSize(2);

        assertEquals(selector1.getName(), result.get(0).getName());
        assertEquals(selector1.getCount(), result.get(0).getCount());
        assertEquals(selector1.getClause().value, result.get(0).getClauseValue());
        assertEquals(selector1.getAttribute(), result.get(0).getAttribute());
        assertEquals(selector1.getFilter(), result.get(0).getFilter());

        assertEquals(selector2.getName(), result.get(1).getName());
        assertEquals(selector2.getCount(), result.get(1).getCount());
        assertEquals(selector2.getClause().value, result.get(1).getClauseValue());
        assertEquals(selector2.getAttribute(), result.get(1).getAttribute());
        assertEquals(selector2.getFilter(), result.get(1).getFilter());
    }

    @Test
    void toGrpcMessages_null() {
        //When + Then
        assertEquals(Collections.emptyList(), SelectorMapper.toGrpcMessages(null));
        assertEquals(Collections.emptyList(), SelectorMapper.toGrpcMessages(new Selector[]{}));
    }

    @ParameterizedTest
    @EnumSource(value = SelectorClause.class)
    void toGrpcMessage_success(SelectorClause clause) {
        //Given
        var selector = new Selector("name", 1, clause, "attribute", "filter");

        //When
        var result = SelectorMapper.toGrpcMessage(selector);

        //Then
        assertNotNull(result);
        assertEquals(selector.getName(), result.getName());
        assertEquals(selector.getCount(), result.getCount());
        assertEquals(selector.getClause().value, result.getClauseValue());
        assertEquals(selector.getAttribute(), result.getAttribute());
        assertEquals(selector.getFilter(), result.getFilter());
    }

    @Test
    void toGrpcMessage_null() {
        //When + Then
        assertThrows(ValidationFrostFSException.class, () -> SelectorMapper.toGrpcMessage(null));
    }

    @Test
    void toGrpcMessage_notValidOperation() {
        //Given
        var selector = new Selector("name", 1, SelectorClause.SAME, "attribute", "filter");


        //When + Then
        try (MockedStatic<Types.Clause> mockStatic = mockStatic(Types.Clause.class)) {
            mockStatic.when(() -> Types.Clause.forNumber(selector.getClause().value))
                    .thenReturn(null);

            assertThrows(ProcessFrostFSException.class, () -> SelectorMapper.toGrpcMessage(selector));
        }
    }

    @ParameterizedTest
    @EnumSource(value = Types.Clause.class, names = "UNRECOGNIZED", mode = EnumSource.Mode.EXCLUDE)
    void toModels_success(Types.Clause clause) {
        //Given
        var selector1 = Types.Selector.newBuilder()
                .setName("name1")
                .setCount(1)
                .setClause(clause)
                .setAttribute("attribute1")
                .setFilter("filter1")
                .build();
        var selector2 = Types.Selector.newBuilder()
                .setName("name2")
                .setCount(2)
                .setClause(clause)
                .setAttribute("attribute2")
                .setFilter("filter2")
                .build();

        //When
        var result = SelectorMapper.toModels(List.of(selector1, selector2));

        //Then
        assertThat(result).isNotNull().hasSize(2);


        assertNotNull(result);
        assertEquals(selector1.getName(), result[0].getName());
        assertEquals(selector1.getCount(), result[0].getCount());
        assertEquals(selector1.getClauseValue(), result[0].getClause().value);
        assertEquals(selector1.getAttribute(), result[0].getAttribute());
        assertEquals(selector1.getFilter(), result[0].getFilter());

        assertEquals(selector2.getName(), result[1].getName());
        assertEquals(selector2.getCount(), result[1].getCount());
        assertEquals(selector2.getClauseValue(), result[1].getClause().value);
        assertEquals(selector2.getAttribute(), result[1].getAttribute());
        assertEquals(selector2.getFilter(), result[1].getFilter());
    }

    @Test
    void toModels_null() {
        //When + Then
        assertNull(SelectorMapper.toModels(null));
        assertNull(SelectorMapper.toModels(Collections.emptyList()));
    }

    @ParameterizedTest
    @EnumSource(value = Types.Clause.class, names = "UNRECOGNIZED", mode = EnumSource.Mode.EXCLUDE)
    void toModel_success(Types.Clause clause) {
        //Given
        var selector = Types.Selector.newBuilder()
                .setName("name")
                .setCount(1)
                .setClause(clause)
                .setAttribute("attribute")
                .setFilter("filter")
                .build();

        //When
        var result = SelectorMapper.toModel(selector);

        //Then
        assertNotNull(result);
        assertEquals(selector.getName(), result.getName());
        assertEquals(selector.getCount(), result.getCount());
        assertEquals(selector.getClauseValue(), result.getClause().value);
        assertEquals(selector.getAttribute(), result.getAttribute());
        assertEquals(selector.getFilter(), result.getFilter());
    }

    @Test
    void toModel_null() {
        //When + Then
        assertNull(SelectorMapper.toModel(null));
        assertNull(SelectorMapper.toModel(Types.Selector.getDefaultInstance()));
    }

    @Test
    void toModel_notValidScheme() {
        //Given
        var selector = Types.Selector.newBuilder()
                .setName("name")
                .setCount(1)
                .setClause(Types.Clause.SAME)
                .setAttribute("attribute")
                .setFilter("filter")
                .build();

        //When + Then
        try (MockedStatic<SelectorClause> mockStatic = mockStatic(SelectorClause.class)) {
            mockStatic.when(() -> SelectorClause.get(Types.Clause.SAME.getNumber()))
                    .thenReturn(null);

            assertThrows(ProcessFrostFSException.class, () -> SelectorMapper.toModel(selector));
        }
    }
}
